import ast
from typing import Generator
from typing import Iterator

from inline_snapshot._customize._custom_dict import CustomDict
from inline_snapshot._customize._custom_undefined import CustomUndefined

from .._adapter_context import AdapterContext
from .._change import ChangeBase
from .._change import Delete
from .._change import DictInsert
from .._global_state import state
from .generic_value import GenericValue
from .undecided_value import UndecidedValue


class DictValue(GenericValue):
    _current_op = "snapshot[key]"

    _new_value: CustomDict
    _old_value: CustomDict
    _ast_node: ast.Dict

    def __getitem__(self, index):
        if isinstance(self._new_value, CustomUndefined):
            self._new_value = CustomDict({})

        index = self.to_custom(index)

        if index not in self._new_value.value:
            if isinstance(self._old_value, CustomUndefined):
                state().missing_values += 1
                old_value = {}
            else:
                old_value = self._old_value.value

            child_node = None
            if self._ast_node is not None:
                assert isinstance(self._ast_node, ast.Dict)
                old_keys = [k for k in old_value.keys()]
                if index in old_keys:
                    pos = old_keys.index(index)
                    child_node = self._ast_node.values[pos]

            self._new_value.value[index] = UndecidedValue(
                old_value.get(index, CustomUndefined()), child_node, self._context
            )

        return self._new_value.value[index]

    def _re_eval(self, value, context: AdapterContext):
        super()._re_eval(value, context)

        if not isinstance(self._new_value, CustomUndefined) and not isinstance(
            self._old_value, CustomUndefined
        ):
            for key, s in self._new_value.value.items():
                if key in self._old_value.value:
                    s._re_eval(self._old_value.value[key], context)  # type:ignore

    def _new_code(self) -> Generator[ChangeBase, None, str]:
        values = []
        for k, v in self._new_value.value.items():
            if not isinstance(v, UndecidedValue):
                new_code = yield from v._new_code()  # type:ignore
                new_key = yield from k._code_repr(self._context)
                values.append(f"{new_key}: {new_code}")

        return "{" + ", ".join(values) + "}"

    def _get_changes(self) -> Iterator[ChangeBase]:

        assert not isinstance(self._old_value, CustomUndefined)

        if self._ast_node is None:
            values = [None] * len(self._old_value.value)
        else:
            assert isinstance(self._ast_node, ast.Dict)
            values = self._ast_node.values

        for key, node in zip(self._old_value.value.keys(), values):
            if key in self._new_value.value:
                # check values with same keys
                yield from self._new_value.value[key]._get_changes()  # type:ignore
            else:
                # delete entries
                yield Delete("trim", self._file, node, self._old_value.value[key])

        to_insert = []
        to_insert_values = []
        for key, new_value_element in self._new_value.value.items():
            if key not in self._old_value.value and not isinstance(
                new_value_element, UndecidedValue
            ):
                # add new values
                new_value = yield from new_value_element._new_code()  # type:ignore
                new_key = yield from key._code_repr(self._context)

                to_insert.append((new_key, new_value))
                to_insert_values.append((key, new_value_element))

        if to_insert:
            yield DictInsert(
                "create",
                self._file,
                self._ast_node,
                len(self._old_value.value),
                to_insert,
                to_insert_values,
            )



from ._binary import BinaryFormat
from ._json import JsonFormat
from ._protocol import Format
from ._protocol import get_format_handler
from ._protocol import get_format_handler_from_suffix
from ._protocol import register_format
from ._protocol import register_format_alias
from ._text import TextFormat

__all__ = (
    "get_format_handler_from_suffix",
    "get_format_handler",
    "Format",
    "register_format_alias",
    "register_format",
    "TextFormat",
    "BinaryFormat",
    "JsonFormat",
)



from ._types import SnapshotBase

is_fixed = False


def pydantic_fix():
    global is_fixed
    if is_fixed:
        return  # pragma: no cover
    is_fixed = True

    try:
        from pydantic import BaseModel
    except ImportError:  # pragma: no cover
        return

    import pydantic

    if not pydantic.version.VERSION.startswith("1."):
        return

    origin_eq = BaseModel.__eq__

    def new_eq(self, other):
        if isinstance(other, SnapshotBase):  # type: ignore
            return other == self
        else:
            return origin_eq(self, other)

    BaseModel.__eq__ = new_eq



from itertools import groupby


def align(seq_a, seq_b) -> str:

    start = 0

    for a, b in zip(seq_a, seq_b):
        if a == b:
            start += 1
        else:
            break

    if start == len(seq_a) == len(seq_b):
        return "m" * start

    end = 0

    for a, b in zip(reversed(seq_a[start:]), reversed(seq_b[start:])):
        if a == b:
            end += 1
        else:
            break

    diff = nw_align(seq_a[start : len(seq_a) - end], seq_b[start : len(seq_b) - end])

    return "m" * start + diff + "m" * end


def nw_align(seq_a, seq_b) -> str:

    matrix: list = [[(0, "e")] + [(0, "i")] * len(seq_b)]

    for a in seq_a:
        last = matrix[-1]

        new_line = [(0, "d")]
        for bi, b in enumerate(seq_b, 1):
            la, lc, lb = new_line[-1], last[bi - 1], last[bi]
            values = [(la[0], "i"), (lb[0], "d")]
            if a == b:
                values.append((lc[0] + 1, "m"))

            new_line.append(max(values))
        matrix.append(new_line)

    # backtrack

    ai = len(seq_a)
    bi = len(seq_b)
    d = ""
    track = ""

    while d != "e":
        _, d = matrix[ai][bi]
        if d == "m":
            ai -= 1
            bi -= 1
        elif d == "i":
            bi -= 1
        elif d == "d":
            ai -= 1
        if d != "e":
            track += d

    return track[::-1]


def add_x(track):
    """Replaces an `id` with the same number of insertions and deletions with
    x."""
    groups = [(c, len(list(v))) for c, v in groupby(track)]
    i = 0
    result = ""
    while i < len(groups):
        g = groups[i]
        if i == len(groups) - 1:
            result += g[0] * g[1]
            break

        ng = groups[i + 1]
        if g[0] == "d" and ng[0] == "i" and g[1] == ng[1]:
            result += "x" * g[1]
            i += 1
        else:
            result += g[0] * g[1]

        i += 1

    return result



from inline_snapshot._customize._custom_code import Import
from inline_snapshot._customize._custom_code import ImportFrom

from .._customize._builder import Builder
from .._customize._custom import Custom
from ._spec import InlineSnapshotPluginSpec
from ._spec import customize
from ._spec import hookimpl

__all__ = (
    "InlineSnapshotPluginSpec",
    "customize",
    "hookimpl",
    "Builder",
    "Custom",
    "Import",
    "ImportFrom",
)



from __future__ import annotations

import ast
from dataclasses import dataclass
from dataclasses import field
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase

from ._custom import Custom


@dataclass(frozen=True)
class CustomDict(Custom):
    node_type = ast.Dict
    value: dict[Custom, Custom] = field(compare=False)

    def _map(self, f):
        return f({k._map(f): v._map(f) for k, v in self.value.items()})

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        values = []
        for k, v in self.value.items():
            key = yield from k._code_repr(context)
            value = yield from v._code_repr(context)
            values.append(f"{key}: {value}")

        return f"{{{ ', '.join(values)}}}"



from __future__ import annotations

import ast
import warnings
from collections import defaultdict
from typing import Generator
from typing import Sequence

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._align import add_x
from inline_snapshot._align import align
from inline_snapshot._change import CallArg
from inline_snapshot._change import ChangeBase
from inline_snapshot._change import Delete
from inline_snapshot._change import DictInsert
from inline_snapshot._change import ListInsert
from inline_snapshot._change import Replace
from inline_snapshot._compare_context import compare_context
from inline_snapshot._customize._custom import Custom
from inline_snapshot._customize._custom_call import CustomCall
from inline_snapshot._customize._custom_call import CustomDefault
from inline_snapshot._customize._custom_code import CustomCode
from inline_snapshot._customize._custom_dict import CustomDict
from inline_snapshot._customize._custom_sequence import CustomList
from inline_snapshot._customize._custom_sequence import CustomSequence
from inline_snapshot._customize._custom_sequence import CustomTuple
from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._customize._custom_unmanaged import CustomUnmanaged
from inline_snapshot._exceptions import UsageError
from inline_snapshot._generator_utils import make_gen_map
from inline_snapshot._generator_utils import only_value
from inline_snapshot._generator_utils import split_gen
from inline_snapshot.syntax_warnings import InlineSnapshotInfo
from inline_snapshot.syntax_warnings import InlineSnapshotSyntaxWarning


def warn_star_expression(node, context):
    if isinstance(node, ast.Call):
        for pos_arg in node.args:
            if isinstance(pos_arg, ast.Starred):
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=context.file._source.filename,
                    lineno=pos_arg.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return True

        # keyword arguments
        for kw in node.keywords:
            if kw.arg is None:
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=context.file._source.filename,
                    lineno=kw.value.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return True

    if isinstance(node, (ast.Tuple, ast.List)):

        for e in node.elts:
            if isinstance(e, ast.Starred):
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=context.file.filename,
                    lineno=e.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return True
    if isinstance(node, ast.Dict):

        for key1, value in zip(node.keys, node.values):
            if key1 is None:
                warnings.warn_explicit(
                    "star-expressions are not supported inside snapshots",
                    filename=context.file._source.filename,
                    lineno=value.lineno,
                    category=InlineSnapshotSyntaxWarning,
                )
                return True

    return False


def reeval(old_value: Custom, value: Custom) -> Custom:
    function_name = f"reeval_{type(old_value).__name__}"
    result = globals()[function_name](old_value, value)
    assert isinstance(result, Custom)

    return result


def reeval_CustomList(old_value: CustomList, value: CustomList):
    assert len(old_value.value) == len(value.value)
    return CustomList([reeval(a, b) for a, b in zip(old_value.value, value.value)])


def reeval_CustomTuple(old_value: CustomTuple, value: CustomTuple):
    assert len(old_value.value) == len(value.value)
    return CustomTuple([reeval(a, b) for a, b in zip(old_value.value, value.value)])


def reeval_CustomUnmanaged(old_value: CustomUnmanaged, value: CustomUnmanaged):
    old_value.value = value.value
    return old_value


def reeval_CustomUndefined(old_value, value):
    return value


def reeval_CustomCode(old_value: CustomCode, value: CustomCode):

    if not old_value._eval() == value._eval():
        raise UsageError(
            "snapshot value should not change. Use Is(...) for dynamic snapshot parts."
        )

    return value


def reeval_CustomCall(old_value: CustomCall, value: CustomCall):
    return CustomCall(
        reeval(old_value.function, value.function),
        [reeval(a, b) for a, b in zip(old_value.args, value.args)],
        {k: reeval(old_value.kwargs[k], value.kwargs[k]) for k in old_value.kwargs},
    )


def reeval_CustomDict(old_value, value):
    assert len(old_value.value) == len(value.value)
    return CustomDict(
        {
            reeval(k1, k2): reeval(v1, v2)
            for (k1, v1), (k2, v2) in zip(old_value.value.items(), value.value.items())
        }
    )


class NewAdapter:

    def __init__(self, context: AdapterContext):
        self.context = context

    def compare(
        self, old_value: Custom, old_node, new_value: Custom
    ) -> Generator[ChangeBase, None, Custom]:

        if isinstance(old_value, CustomUnmanaged):
            return old_value

        if isinstance(new_value, CustomUnmanaged):
            raise UsageError("unmanaged values can not be compared with snapshots")

        if (
            type(old_value) is type(new_value)
            and (
                isinstance(old_node, new_value.node_type)
                if old_node is not None
                else True
            )
            and (
                isinstance(old_value, (CustomCall, CustomSequence, CustomDict))
                if old_node is None
                else True
            )
        ):
            function_name = f"compare_{type(old_value).__name__}"
            result = yield from getattr(self, function_name)(
                old_value, old_node, new_value
            )
        else:
            result = yield from self.compare_CustomCode(old_value, old_node, new_value)
        return result

    def compare_CustomCode(
        self, old_value: Custom, old_node: ast.expr, new_value: Custom
    ) -> Generator[ChangeBase, None, Custom]:

        assert isinstance(old_value, Custom)
        assert isinstance(new_value, Custom)
        assert isinstance(old_node, (ast.expr, type(None))), old_node

        new_code, new_changes = split_gen(new_value._code_repr(self.context))

        if (
            isinstance(old_node, ast.JoinedStr)
            and isinstance(new_value, CustomCode)
            and isinstance(new_value.value, str)
        ):
            if not old_value._eval() == new_value._eval():

                value = only_value(new_value._code_repr(self.context))
                warnings.warn_explicit(
                    f"inline-snapshot will be able to fix f-strings in the future.\nThe current string value is:\n   {value}",
                    filename=self.context.file._source.filename,
                    lineno=old_node.lineno,
                    category=InlineSnapshotInfo,
                )
            return old_value

        if not old_value._eval() == new_value.original_value:
            if isinstance(old_value, CustomUndefined):
                flag = "create"
            else:
                flag = "fix"
        elif not isinstance(
            old_value, CustomUnmanaged
        ) and self.context.file.code_changed(old_node, new_code):
            flag = "update"
        else:
            # equal and equal repr
            return old_value

        for change in new_changes:
            change.flag = flag
            yield change

        yield Replace(
            node=old_node,
            file=self.context.file,
            new_code=new_code,
            flag=flag,
            old_value=old_value._eval(),
            new_value=new_value,
        )

        return new_value

    def compare_CustomSequence(
        self, old_value: CustomSequence, old_node: ast.AST, new_value: CustomSequence
    ) -> Generator[ChangeBase, None, CustomSequence]:

        if old_node is not None:
            assert isinstance(
                old_node, ast.List if isinstance(old_value._eval(), list) else ast.Tuple
            )
            assert isinstance(old_node, (ast.List, ast.Tuple))

        else:
            pass  # pragma: no cover

        with compare_context():
            diff = add_x(align(old_value.value, new_value.value))
        old = zip(
            old_value.value,
            old_node.elts if old_node is not None else [None] * len(old_value.value),
        )
        new = iter(new_value.value)
        old_position = 0
        to_insert = defaultdict(list)
        result = []
        for c in diff:
            if c in "mx":
                old_value_element, old_node_element = next(old)
                new_value_element = next(new)
                v = yield from self.compare(
                    old_value_element, old_node_element, new_value_element
                )
                result.append(v)
                old_position += 1
            elif c == "i":
                new_value_element = next(new)
                new_code = yield from new_value_element._code_repr(self.context)
                result.append(new_value_element)
                to_insert[old_position].append((new_code, new_value_element))
            elif c == "d":
                old_value_element, old_node_element = next(old)
                yield Delete(
                    "fix",
                    self.context.file,
                    old_node_element,
                    old_value_element,
                )
                old_position += 1
            else:
                assert False

        for position, code_values in to_insert.items():
            yield ListInsert(
                "fix",
                self.context.file,
                old_node,
                position,
                *zip(*code_values),  # type:ignore
            )

        return type(new_value)(result)

    compare_CustomTuple = compare_CustomSequence
    compare_CustomList = compare_CustomSequence

    def compare_CustomDict(
        self, old_value: CustomDict, old_node: ast.Dict, new_value: CustomDict
    ) -> Generator[ChangeBase, None, Custom]:
        assert isinstance(old_value, CustomDict)
        assert isinstance(new_value, CustomDict)

        if old_node is not None:

            for value2, node in zip(old_value.value.keys(), old_node.keys):
                assert node is not None
                try:
                    # this is just a sanity check, dicts should be ordered
                    node_value = ast.literal_eval(node)
                except Exception:
                    continue
                assert node_value == value2._eval()
        else:
            pass  # pragma: no cover

        result = {}
        for key2, node2 in zip(
            old_value.value.keys(),
            (
                old_node.values
                if old_node is not None
                else [None] * len(old_value.value)
            ),
        ):
            if key2 not in new_value.value:
                # delete entries
                yield Delete("fix", self.context.file, node2, old_value.value[key2])

        to_insert = []
        insert_pos = 0
        for key, new_value_element in new_value.value.items():
            if key not in old_value.value:
                # add new values
                to_insert.append((key, new_value_element))
                result[key] = new_value_element
            else:
                if isinstance(old_node, ast.Dict):
                    node = old_node.values[list(old_value.value.keys()).index(key)]
                else:
                    node = None
                # check values with same keys
                result[key] = yield from self.compare(
                    old_value.value[key], node, new_value.value[key]
                )

                if to_insert:
                    new_code = []
                    for k, v in to_insert:
                        new_code_key = yield from k._code_repr(self.context)
                        new_code_value = yield from v._code_repr(self.context)
                        new_code.append((new_code_key, new_code_value))

                    yield DictInsert(
                        "fix",
                        self.context.file,
                        old_node,
                        insert_pos,
                        new_code,
                        to_insert,
                    )
                    to_insert = []

                insert_pos += 1

        if to_insert:
            new_code = []
            for k, v in to_insert:
                new_code_key = yield from k._code_repr(self.context)
                new_code_value = yield from v._code_repr(self.context)
                new_code.append(
                    (
                        new_code_key,
                        new_code_value,
                    )
                )

            yield DictInsert(
                "fix",
                self.context.file,
                old_node,
                len(old_value.value),
                new_code,
                to_insert,
            )

        return CustomDict(value=result)

    def compare_CustomCall(
        self, old_value: CustomCall, old_node: ast.Call, new_value: CustomCall
    ) -> Generator[ChangeBase, None, Custom]:

        call = new_value
        new_args = call.args
        new_kwargs = call.kwargs

        # positional arguments

        result_args = []

        flag = "update" if old_value._eval() == new_value.original_value else "fix"

        @make_gen_map
        def intercept(change):
            if flag == "update" and change.flag == "fix":
                change.flag = "update"
            return change

        old_node_args: Sequence[ast.expr | None]
        if old_node:
            old_node_args = old_node.args
        else:
            old_node_args = [None] * len(new_args)

        for i, (new_value_element, node) in enumerate(zip(new_args, old_node_args)):
            old_value_element = old_value.argument(i)
            result = yield from intercept(
                self.compare(old_value_element, node, new_value_element)
            )
            result_args.append(result)

        old_args_len = len(old_node.args if old_node else old_value.args)

        if old_node is not None:
            if old_args_len > len(new_args):
                for arg_pos, node in list(enumerate(old_node.args))[len(new_args) :]:
                    yield Delete(
                        flag,
                        self.context.file,
                        node,
                        old_value.argument(arg_pos),
                    )

        if old_args_len < len(new_args):
            for insert_pos, insert_value in list(enumerate(new_args))[old_args_len:]:
                new_code = yield from insert_value._code_repr(self.context)
                yield CallArg(
                    flag=flag,
                    file=self.context.file,
                    node=old_node,
                    arg_pos=insert_pos,
                    arg_name=None,
                    new_code=new_code,
                    new_value=insert_value,
                )
                result_args.append(insert_value)

        # keyword arguments
        result_kwargs = {}
        if old_node is None:
            old_keywords = {key: None for key in old_value.kwargs.keys()}
        else:
            old_keywords = {
                kw.arg: kw.value for kw in old_node.keywords if kw.arg is not None
            }

        for kw_arg, kw_value in old_keywords.items():
            missing = kw_arg not in new_kwargs
            if missing or isinstance(new_kwargs[kw_arg], CustomDefault):
                # delete entries
                yield Delete(
                    (
                        "update"
                        if not missing
                        and old_value.argument(kw_arg) == new_value.argument(kw_arg)
                        else flag
                    ),
                    self.context.file,
                    kw_value,
                    old_value.argument(kw_arg),
                )

        to_insert = []
        insert_pos = len(old_value.args)
        for key, new_value_element in new_kwargs.items():
            if isinstance(new_value_element, CustomDefault):
                continue
            if key not in old_keywords:
                # add new values
                to_insert.append((key, new_value_element))
                result_kwargs[key] = new_value_element
            else:
                node = old_keywords[key]

                # check values with same keys
                old_value_element = old_value.argument(key)
                result_kwargs[key] = yield from intercept(
                    self.compare(old_value_element, node, new_value_element)
                )

                if to_insert:
                    for insert_key, value in to_insert:
                        new_code = yield from value._code_repr(self.context)
                        yield CallArg(
                            flag=flag,
                            file=self.context.file,
                            node=old_node,
                            arg_pos=insert_pos,
                            arg_name=insert_key,
                            new_code=new_code,
                            new_value=value,
                        )
                    to_insert = []

                insert_pos += 1

        if to_insert:

            for key, value in to_insert:
                new_code = yield from value._code_repr(self.context)

                yield CallArg(
                    flag=flag,
                    file=self.context.file,
                    node=old_node,
                    arg_pos=None,
                    arg_name=key,
                    new_code=new_code,
                    new_value=value,
                )

        return CustomCall(
            (
                yield from intercept(
                    self.compare(
                        old_value.function,
                        old_node.func if old_node else None,
                        new_value.function,
                    )
                )
            ),
            result_args,
            result_kwargs,
        )



import ast
import datetime
from collections import Counter
from collections import defaultdict
from dataclasses import MISSING
from dataclasses import fields
from dataclasses import is_dataclass
from enum import Enum
from enum import Flag
from pathlib import Path
from pathlib import PurePath
from types import BuiltinFunctionType
from types import FunctionType
from typing import Any
from typing import Dict

from inline_snapshot._customize._builder import Builder
from inline_snapshot._customize._custom_code import ImportFrom
from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._customize._custom_unmanaged import CustomUnmanaged
from inline_snapshot._external._outsource import Outsourced
from inline_snapshot._sentinels import undefined
from inline_snapshot._unmanaged import is_dirty_equal
from inline_snapshot._unmanaged import is_unmanaged
from inline_snapshot._utils import triple_quote

from ._spec import customize


class InlineSnapshotPlugin:
    @customize
    def standard_handler(self, value, builder: Builder):
        if isinstance(value, list):
            return builder.create_list(value)

        if type(value) is tuple:
            return builder.create_tuple(value)

        if isinstance(value, dict):
            return builder.create_dict(value)

    @customize
    def string_handler(self, value, builder: Builder):
        if isinstance(value, str) and (
            ("\n" in value and value[-1] != "\n") or value.count("\n") > 1
        ):

            triple_quoted_string = triple_quote(value)

            assert ast.literal_eval(triple_quoted_string) == value

            return builder.create_code(triple_quoted_string)

    @customize(tryfirst=True)
    def counter_handler(self, value, builder: Builder):
        if isinstance(value, Counter):
            return builder.create_call(Counter, [dict(value)])

    @customize
    def function_and_type_handler(
        self, value, builder: Builder, local_vars: Dict[str, Any]
    ):
        if isinstance(value, (FunctionType, type)):
            for name, local_value in local_vars.items():
                if local_value is value:
                    return builder.create_code(name)

            qualname = value.__qualname__.split("[")[0]
            name = qualname.split(".")[0]
            return builder.create_code(
                qualname, imports=[ImportFrom(value.__module__, name)]
            )

    @customize
    def builtin_function_handler(self, value, builder: Builder):
        if isinstance(value, BuiltinFunctionType):
            return builder.create_code(value.__name__)

    @customize
    def timezone_handler(self, value, builder: Builder):
        if type(value) is datetime.timezone:
            # Handle timezone.utc specially - it's a constant, not a constructor call
            if value == datetime.timezone.utc:
                return builder.create_code(
                    "timezone.utc",
                    imports=[ImportFrom("datetime", "timezone")],
                )

            # For other timezone objects, use the constructor
            offset = value.utcoffset(None)
            tzname = value.tzname(None)
            return builder.create_call(datetime.timezone, [offset, tzname])

    @customize
    def datetime_handler(self, value, builder: Builder):

        if type(value) is datetime.datetime:
            return builder.create_call(
                datetime.datetime,
                [value.year, value.month, value.day],
                {
                    "hour": builder.with_default(value.hour, 0),
                    "minute": builder.with_default(value.minute, 0),
                    "second": builder.with_default(value.second, 0),
                    "microsecond": builder.with_default(value.microsecond, 0),
                    "tzinfo": builder.with_default(value.tzinfo, None),
                },
            )

        if type(value) is datetime.date:
            return builder.create_call(
                datetime.date, [value.year, value.month, value.day]
            )

        if type(value) is datetime.time:
            return builder.create_call(
                datetime.time,
                [],
                {
                    "hour": builder.with_default(value.hour, 0),
                    "minute": builder.with_default(value.minute, 0),
                    "second": builder.with_default(value.second, 0),
                    "microsecond": builder.with_default(value.microsecond, 0),
                    "tzinfo": builder.with_default(value.tzinfo, None),
                },
            )

        if type(value) is datetime.timedelta:
            return builder.create_call(
                datetime.timedelta,
                [],
                {
                    "days": builder.with_default(value.days, 0),
                    "seconds": builder.with_default(value.seconds, 0),
                    "microseconds": builder.with_default(value.microseconds, 0),
                },
            )

    @customize
    def path_handler(self, value, builder: Builder):
        if isinstance(value, Path):
            return builder.create_call(Path, [value.as_posix()])

        if isinstance(value, PurePath):
            return builder.create_call(PurePath, [value.as_posix()])

    def sort_set_values(self, set_values):
        is_sorted = False
        try:
            set_values = sorted(set_values)
            is_sorted = True
        except TypeError:
            # can not be sorted by value
            pass

        set_values = list(map(repr, set_values))
        if not is_sorted:
            # sort by string representation
            set_values = sorted(set_values)

        return set_values

    @customize
    def set_handler(self, value, builder: Builder):
        if isinstance(value, set):
            if len(value) == 0:
                return builder.create_code("set()")
            else:
                return builder.create_code(
                    "{" + ", ".join(self.sort_set_values(value)) + "}"
                )

    @customize
    def frozenset_handler(self, value, builder: Builder):
        if isinstance(value, frozenset):
            if len(value) == 0:
                return builder.create_code("frozenset()")
            else:
                return builder.create_call(frozenset, [set(value)])

    # -8<- [start:Enum]
    @customize
    def enum_handler(self, value, builder: Builder):
        if isinstance(value, Enum):
            qualname = type(value).__qualname__
            name = qualname.split(".")[0]

            return builder.create_code(
                f"{type(value).__qualname__}.{value.name}",
                imports=[ImportFrom(type(value).__module__, name)],
            )

    # -8<- [end:Enum]

    @customize
    def flag_handler(self, value, builder: Builder):
        if isinstance(value, Flag):
            qualname = type(value).__qualname__
            name = qualname.split(".")[0]

            return builder.create_code(
                " | ".join(
                    f"{qualname}.{flag.name}" for flag in type(value) if flag in value
                ),
                imports=[ImportFrom(type(value).__module__, name)],
            )

    @customize
    def source_file_name_handler(self, value, builder: Builder, global_vars):
        if "__file__" in global_vars and value == global_vars["__file__"]:
            return builder.create_code("__file__")

    @customize
    def dataclass_handler(self, value, builder: Builder):

        if is_dataclass(value) and not isinstance(value, type):

            kwargs = {}

            for field in fields(value):  # type: ignore
                if field.repr:
                    field_value = getattr(value, field.name)

                    if field.default != MISSING:
                        field_value = builder.with_default(field_value, field.default)

                    if field.default_factory != MISSING:
                        field_value = builder.with_default(
                            field_value, field.default_factory()
                        )

                    kwargs[field.name] = field_value

            return builder.create_call(type(value), [], kwargs)

    @customize
    def namedtuple_handler(self, value, builder: Builder):
        t = type(value)
        b = t.__bases__
        if len(b) != 1 or b[0] != tuple:
            return
        f = getattr(t, "_fields", None)
        if not isinstance(f, tuple):
            return
        if not all(type(n) == str for n in f):
            return

        return builder.create_call(
            type(value),
            [],
            {
                field: (
                    getattr(value, field)
                    if field not in value._field_defaults
                    else builder.with_default(
                        getattr(value, field), value._field_defaults[field]
                    )
                )
                for field in value._fields
            },
        )

    @customize(tryfirst=True)
    def defaultdict_handler(self, value, builder: Builder):
        if isinstance(value, defaultdict):
            return builder.create_call(
                type(value), [value.default_factory, dict(value)], {}
            )

    @customize
    def unmanaged_handler(self, value, builder: Builder):
        if is_unmanaged(value):
            return CustomUnmanaged(value=value)

    @customize
    def undefined_handler(self, value, builder: Builder):
        if value is undefined:
            return CustomUndefined()

    @customize
    def outsource_handler(self, value, builder: Builder):
        if isinstance(value, Outsourced):
            return builder.create_external(
                value.data, format=value.suffix, storage=value.storage
            )


try:
    import dirty_equals
except ImportError:  # pragma: no cover
    pass
else:

    from dirty_equals._utils import Omit

    class InlineSnapshotDirtyEqualsPlugin:
        @customize(tryfirst=True)
        def dirty_equals_handler(self, value, builder: Builder):

            if is_dirty_equal(value) and builder._build_new_value:

                if isinstance(value, type):
                    return builder.create_code(
                        value.__name__,
                        imports=[ImportFrom("dirty_equals", value.__name__)],
                    )
                else:

                    args = [a for a in value._repr_args if a is not Omit]
                    kwargs = {
                        k: a for k, a in value._repr_kwargs.items() if a is not Omit
                    }
                    if type(value) == dirty_equals.IsNow:
                        kwargs.pop("approx")
                        if (
                            isinstance(delta := kwargs["delta"], datetime.timedelta)
                            and delta.total_seconds() == 2
                        ):
                            kwargs.pop("delta")
                    return builder.create_call(type(value), args, kwargs)

        @customize(tryfirst=True)
        def is_now_handler(self, value, builder: Builder):
            if value == dirty_equals.IsNow():
                return dirty_equals.IsNow()


try:
    import attrs
except ImportError:  # pragma: no cover

    pass

else:

    class InlineSnapshotAttrsPlugin:
        @customize
        def attrs_handler(self, value, builder: Builder):

            if attrs.has(type(value)):

                kwargs = {}

                for field in attrs.fields(type(value)):
                    if field.repr:
                        field_value = getattr(value, field.name)

                        if field.default is not attrs.NOTHING:

                            default_value = (
                                field.default
                                if not isinstance(field.default, attrs.Factory)  # type: ignore
                                else (
                                    field.default.factory()
                                    if not field.default.takes_self
                                    else field.default.factory(value)
                                )
                            )
                            field_value = builder.with_default(
                                field_value, default_value
                            )

                        kwargs[field.name] = field_value

                return builder.create_call(type(value), [], kwargs)


try:
    import pydantic
except ImportError:  # pragma: no cover

    pass

else:
    # import pydantic
    if pydantic.version.VERSION.startswith("1."):
        # pydantic v1
        from pydantic.fields import Undefined as PydanticUndefined  # type: ignore[attr-defined,no-redef]

        def get_fields(value):
            return value.__fields__

    else:
        # pydantic v2
        from pydantic_core import PydanticUndefined

        def get_fields(value):
            return type(value).model_fields

    from pydantic import BaseModel

    class InlineSnapshotPydanticPlugin:
        @customize
        def pydantic_model_handler(self, value, builder: Builder):

            if isinstance(value, BaseModel):

                kwargs = {}

                for name, field in get_fields(value).items():  # type: ignore
                    if getattr(field, "repr", True):
                        field_value = getattr(value, name)

                        if (
                            field.default is not PydanticUndefined
                            and field.default == field_value
                        ):
                            field_value = builder.with_default(
                                field_value, field.default
                            )

                        elif field.default_factory is not None:
                            field_value = builder.with_default(
                                field_value, field.default_factory()
                            )

                        kwargs[name] = field_value

                return builder.create_call(type(value), [], kwargs)



import os
import sys
import tokenize
from pathlib import Path
from types import FunctionType
from types import SimpleNamespace
from typing import Dict
from typing import List

from executing import is_pytest_compatible
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm
from rich.syntax import Syntax

from inline_snapshot._external._storage import default_storages
from inline_snapshot._utils import category_link
from inline_snapshot._utils import link

from . import _config
from ._change import ChangeBase
from ._change import ExternalRemove
from ._change import apply_all
from ._external._find_external import used_externals_in
from ._flags import Flags
from ._global_state import state
from ._problems import report_problems
from ._rewrite_code import ChangeRecorder
from .fix_pytest_diff import fix_pytest_diff
from .pydantic_fix import pydantic_fix
from .version import is_insider

if sys.version_info >= (3, 13):
    # fixes #186
    try:
        import readline  # noqa
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        # should fix #189 and #245
        pass


categories = Flags.all().to_set()


def is_ci_run():
    if bool(os.environ.get("PYCHARM_HOSTED", False)):
        # pycharm exports TEAMCITY_VERSION but luckily also PYCHARM_HOSTED,
        # which allows to detect the incorrect ci detection
        return False

    ci_env_vars = (
        "CI",
        "bamboo.buildKey",
        "BUILD_ID",
        "BUILD_NUMBER",
        "BUILDKITE",
        "CIRCLECI",
        "CONTINUOUS_INTEGRATION",
        "GITHUB_ACTIONS",
        "HUDSON_URL",
        "JENKINS_URL",
        "TEAMCITY_VERSION",
        "TRAVIS",
    )
    for var in ci_env_vars:
        if os.environ.get(var, False):
            return var
    return False


def is_implementation_supported():
    return sys.implementation.name == "cpython"


def call_once(f):
    called = False
    result = None

    def w():
        nonlocal called
        nonlocal result
        if not called:
            result = f()
            called = True
        return result

    return w


def apply_changes(flag, console):
    if flag in state().flags:
        console().print(
            f"These changes will be applied, because you used {category_link(flag)}",
            highlight=False,
        )
        console().print()
        return True
    if "review" in state().flags:
        if is_insider and flag == "fix-assert":
            result = Confirm.ask(
                f"Do you want to {link('fix','https://15r10nk.github.io/inline-snapshot/fix_assert/')} these assertions?",
                default=False,
                console=console(),
            )
        else:
            result = Confirm.ask(
                f"Do you want to {category_link(flag)} these snapshots?",
                default=False,
                console=console(),
            )

        console().print()
        return result
    else:
        console().print("These changes are not applied.")
        console().print(
            f"Use [bold]--inline-snapshot={category_link(flag)} to apply them, "
            "or use the interactive mode with [b]--inline-snapshot=[italic blue link=https://15r10nk.github.io/inline-snapshot/latest/pytest/#-inline-snapshotreview]review[/][/]",
            highlight=False,
        )
        console().print()
        return False


def short_report(snapshot_changes, console):
    def report(flag, message, message_n):
        num = snapshot_changes[flag]

        if num and not getattr(state().update_flags, flag):
            console().print(
                message if num == 1 else message_n.format(num=num),
                highlight=False,
            )

    report(
        "fix",
        "Error: one snapshot has incorrect values ([b]--inline-snapshot=fix[/])",
        "Error: {num} snapshots have incorrect values ([b]--inline-snapshot=fix[/])",
    )

    report(
        "trim",
        "Info: one snapshot can be trimmed ([b]--inline-snapshot=trim[/])",
        "Info: {num} snapshots can be trimmed ([b]--inline-snapshot=trim[/])",
    )

    report(
        "create",
        "Error: one snapshot is missing a value ([b]--inline-snapshot=create[/])",
        "Error: {num} snapshots are missing values ([b]--inline-snapshot=create[/])",
    )

    report(
        "update",
        "Info: one snapshot changed its representation ([b]--inline-snapshot=update[/])",
        "Info: {num} snapshots changed their representation ([b]--inline-snapshot=update[/])",
    )

    if sum(snapshot_changes.values()) != 0:
        console().print(
            "\nYou can also use [b]--inline-snapshot=review[/] to approve the changes interactively",
            highlight=False,
        )

    return


def filter_changes(changes, snapshot_changes, console):

    used_changes = []
    for flag in Flags.all():
        if not changes[flag]:
            continue

        if not {"review", "report", flag} & state().flags:
            continue

        @call_once
        def header():
            if is_insider and flag == "fix-assert":
                caption = "fix-assert"
            else:
                caption = flag.capitalize() + " snapshots"
            console().rule(f"[yellow bold]{caption}")

        if (
            flag == "update"
            and not state().config.show_updates
            and not "update" in state().flags
        ):
            continue

        cr = ChangeRecorder()
        apply_all(used_changes, cr)
        cr.virtual_write()
        apply_all(changes[flag], cr)

        any_changes = False

        for file in cr.files():
            diff = file.diff()
            if diff:
                header()
                name = file.filename.resolve().relative_to(Path.cwd().resolve())
                console().print(
                    Panel(
                        Syntax(diff, "diff", theme="ansi_light", word_wrap=True),
                        title=name.as_posix(),
                        box=(
                            box.ASCII
                            if os.environ.get("TERM", "") == "unknown"
                            else box.ROUNDED
                        ),
                    )
                )
                any_changes = True

        for change in changes[flag]:
            diff = change.rich_diff()
            if diff is not None:
                title, content = diff
                console().print(
                    Panel(
                        content,
                        title=title,
                        box=(
                            box.ASCII
                            if os.environ.get("TERM", "") == "unknown"
                            else box.ROUNDED
                        ),
                    )
                )
                any_changes = True

        if any_changes and apply_changes(flag, console):
            used_changes += changes[flag]

    return used_changes


class SnapshotSession:

    def __init__(self):
        self.registered_modules = set()

    def register_customize_hooks_from_module(self, module):
        """Find and register functions decorated with @customize from a module"""

        if module.__file__ in self.registered_modules:
            return

        self.registered_modules.add(module.__file__)

        hooks = {}

        for name in dir(module):
            obj = getattr(module, name, None)
            if isinstance(obj, type) and name.startswith("InlineSnapshot"):
                state().pm.register(obj(), name=f"<conftest {name} {module.__file__}>")

            if isinstance(obj, FunctionType) and hasattr(obj, "inline_snapshot_impl"):
                hooks[name] = obj

        if hooks:
            state().pm.register(
                SimpleNamespace(**hooks), name=f"<conftest {module.__file__}>"
            )

    @staticmethod
    def test_enter():
        state().missing_values = 0
        state().incorrect_values = 0

    @staticmethod
    def test_exit(fail):
        missing_values = state().missing_values
        incorrect_values = state().incorrect_values

        extra = "\nIf you just created this value with --inline-snapshot=create, the value is now created and you can ignore this message."

        if missing_values != 0:
            fail(
                (
                    "your snapshot is missing one value."
                    if missing_values == 1
                    else f"your snapshot is missing {missing_values} values."
                )
                + extra,
            )

        if incorrect_values != 0:
            fail(
                "some snapshots in this test have incorrect values." + extra,
            )

    def load_config(self, pyproject, cli_flags, parallel_run, error, project_root):

        # read config
        if pyproject is not None:
            _config.read_config(pyproject, state().config)

        if state().config.test_directories is None:
            if (tests_dir := Path.cwd() / "tests").exists() and tests_dir.is_dir():
                state().config.test_directories = [tests_dir.resolve()]

        console = Console()

        if is_ci_run():
            default_flags = {"disable"}
            state().disable_reason = "ci"
        elif console.is_terminal:
            default_flags = state().config.default_flags_tui
        else:
            default_flags = state().config.default_flags

        # load flags from env var
        env_var = "INLINE_SNAPSHOT_DEFAULT_FLAGS"
        if env_var in os.environ:
            default_flags = os.environ[env_var].split(",")

        # load flags from command line argument
        if cli_flags is None:
            flags = set(default_flags)
        else:
            state().disable_reason = None
            flags = cli_flags
            if parallel_run and flags - {"disable"}:
                error(
                    f"--inline-snapshot={','.join(flags)} can not be combined with xdist"
                )

        state().flags = flags

        # check flags
        unknown_flags = (
            flags - categories - {"disable", "review", "report", "short-report"}
        )

        if unknown_flags:
            error(
                f"--inline-snapshot={','.join(sorted(unknown_flags))} is a unknown flag"
            )

        if "disable" in flags and flags != {"disable"}:
            error(
                f"--inline-snapshot=disable can not be combined with other flags ({', '.join(flags-{'disable'})})"
            )

        # disable inline-snapshot if it can not be used in the current context
        if parallel_run and "disable" not in flags:
            state().active = False
            state().disable_reason = "xdist"
        elif not is_implementation_supported() and "disable" not in flags:
            state().active = False
            state().disable_reason = "implementation"
        elif flags & {"review"}:
            state().active = True
            state().update_flags = Flags.all()
        else:
            state().active = "disable" not in flags
            state().update_flags = Flags(flags & categories)

        storage_dir = state().config.storage_dir
        if storage_dir is None:
            storage_dir = project_root / ".inline-snapshot"
            state().config.storage_dir = storage_dir

        state().all_storages = default_storages(storage_dir)

        if flags - {"short-report", "disable"} and not is_pytest_compatible():

            # hack to disable the assertion rewriting
            # I found no other way because the hook gets installed early
            sys.meta_path = [
                e for e in sys.meta_path if type(e).__name__ != "AssertionRewritingHook"
            ]

        self.fix_libraries()

    def fix_libraries(self):
        pydantic_fix()
        fix_pytest_diff()

    def show_report(self, con: Console):

        @call_once
        def console():
            con.print("\n")
            con.rule(
                "[blue]inline-snapshot" + (" (insider version)" if is_insider else ""),
                characters="═",
            )

            return con

        if not state().active:
            disable_info = "This means that tests with snapshots will continue to run, but snapshot(x) will only return x and inline-snapshot will not be able to fix snapshots or generate reports."
            if state().disable_reason == "xdist":
                console().print(
                    f"INFO: inline-snapshot was disabled because you used xdist. {disable_info}\n"
                )

            elif state().disable_reason == "ci":
                env_var = is_ci_run()
                console().print(
                    f'INFO: CI run was detected because environment variable "{env_var}" was defined. '
                    f"inline-snapshot runs with --inline-snapshot=disable by default in CI. {disable_info}"
                    " You can change this by using --inline-snapshot=report for example.\n"
                )

            elif state().disable_reason == "implementation":
                console().print(
                    f"INFO: inline-snapshot was disabled because {sys.implementation.name} is not supported. {disable_info}\n"
                )

            return

        # --inline-snapshot

        # auto mode
        changes: Dict[str, List[ChangeBase]] = {f: [] for f in Flags.all()}

        snapshot_changes = {f: 0 for f in Flags.all()}

        for snapshot in state().snapshots.values():
            all_categories = set()
            for change in snapshot._changes():
                changes[change.flag].append(change)
                all_categories.add(change.flag)

            for category in all_categories:
                snapshot_changes[category] += 1

        if "short-report" in state().flags:
            short_report(snapshot_changes, console)
            return

        used_changes = filter_changes(changes, snapshot_changes, console)

        cr = ChangeRecorder()
        apply_all(used_changes, cr)
        changed_files = {Path(f.filename): f for f in cr.files()}

        tests_dir = state().config.test_directories
        if tests_dir:
            assert isinstance(tests_dir, list), tests_dir
            all_files = {
                file for test_dir in tests_dir for file in test_dir.rglob("*.py")
            }

            used = []

            for file in all_files:
                if file in changed_files:
                    content = changed_files[file].new_code()
                    check_import = False
                else:
                    with tokenize.open(file) as f:
                        content = f.read()
                    check_import = True

                used += used_externals_in(file, content, check_import=check_import)

            changes = {f: [] for f in Flags.all()}

            for name, storage in state().all_storages.items():
                externals_for_storage = [e for e in used if e.storage == name]
                for location in storage.find_unused_externals(externals_for_storage):
                    if state().update_flags.trim:
                        changes["trim"].append(ExternalRemove("trim", location))

                storage.check_externals(externals_for_storage)

            used_changes += filter_changes(changes, snapshot_changes, console)

        report_problems(console)

        if used_changes:
            cr = ChangeRecorder()
            apply_all(used_changes, cr)

            for change in used_changes:
                change.apply_external_changes()

            cr.fix_all()



from __future__ import annotations

from pathlib import Path

from inline_snapshot._external._diff import TextDiff

from ._protocol import Format
from ._protocol import register_format


@register_format
class TextFormat(TextDiff, Format[str]):
    "Stores strings in `.txt` files."

    suffix = ".txt"

    def is_format_for(self, value: object):
        return isinstance(value, str)

    def encode(self, value: str, path: Path):
        with path.open("w", encoding="utf-8", newline="\n") as f:
            f.write(value)

    def decode(self, path: Path) -> str:
        with path.open("r", encoding="utf-8", newline="\n") as f:
            return f.read()



from functools import partial
from typing import Any
from typing import Dict

import pluggy

from inline_snapshot._customize._builder import Builder

inline_snapshot_plugin_name = "inline_snapshot"

hookspec = pluggy.HookspecMarker(inline_snapshot_plugin_name)
"""
The pluggy hookspec marker for inline_snapshot.
"""

hookimpl = pluggy.HookimplMarker(inline_snapshot_plugin_name)
"""
The pluggy hookimpl marker for inline_snapshot.
"""

customize = partial(hookimpl, specname="customize")
"""
Decorator to mark a function as an implementation of the `customize` hook which can be used instead of `hookimpl(specname="customize")`.
"""


class InlineSnapshotPluginSpec:
    @hookspec(firstresult=True)
    def customize(
        self,
        value: Any,
        builder: Builder,
        local_vars: Dict[str, Any],
        global_vars: Dict[str, Any],
    ) -> Any:
        """
        The customize hook is called every time a snapshot value should be converted into code.

        This hook allows you to control how inline-snapshot represents objects in generated code.
        When multiple handlers are registered, they are called until one returns a non-None value.
        `customize` is also called for each attribute of the converted hook which is not a Custom node, which means that a hook for `int` does not only apply for `snapshot(5)` but also for `snaspshot([1,2,3])` 3 times.

        Arguments:
            value: The Python object that needs to be converted into source code representation.
                   This is the actual runtime value from your test.
            builder: A Builder instance providing methods to construct custom code representations.
                    Use methods like `create_call()`, `create_dict()`, `create_external()`, etc.
            local_vars: Dictionary mapping variable names to their values in the local scope.
                       Useful for referencing existing variables instead of creating new literals.
            global_vars: Dictionary mapping variable names to their values in the global scope.

        Returns:
            (Custom): created using [Builder][inline_snapshot.plugin.Builder] `create_*` methods.
            (None): if this handler doesn't apply to the given value.
            (Something else): when the next handler should process the value.

        Example:
            You can use @customize when you want to specify multiple handlers in the same class:


            === "with @customize"
                <!-- inline-snapshot-lib: conftest.py -->
                ``` python title="conftest.py"
                from inline_snapshot.plugin import customize


                class InlineSnapshotPlugin:
                    @customize
                    def binary_numbers(self, value, builder, local_vars, global_vars):
                        if isinstance(value, int):
                            return builder.create_code(bin(value))

                    @customize
                    def repeated_strings(self, value, builder):
                        if isinstance(value, str) and value == value[0] * len(value):
                            return builder.create_code(f"'{value[0]}'*{len(value)}")
                ```

            === "by method name"

                <!-- inline-snapshot-lib: conftest.py -->
                ``` python title="conftest.py"
                class InlineSnapshotPlugin:
                    def customize(self, value, builder, local_vars, global_vars):
                        if isinstance(value, int):
                            return builder.create_code(bin(value))

                        if isinstance(value, str) and value == value[0] * len(value):
                            return builder.create_code(f"'{value[0]}'*{len(value)}")
                ```


            <!-- inline-snapshot: create fix first_block outcome-passed=1 -->
            ``` python title="test_customizations.py"
            from inline_snapshot import snapshot


            def test_customizations():
                assert ["aaaaaaaaaaaaaaa", "bbbbbb"] == snapshot(["a" * 15, "b" * 6])
                assert 18856 == snapshot(0b100100110101000)
            ```



        """

    # @hookspec
    # def format_code(self, filename, str) -> str: ...



from __future__ import annotations

import json
from pathlib import Path

from inline_snapshot._external._diff import TextDiff

from ._protocol import Format
from ._protocol import register_format


def is_json(obj):
    if isinstance(obj, (int, str, bool, float, type(None))):
        return True
    if isinstance(obj, dict) and all(
        is_json(v) and isinstance(k, str) for k, v in obj.items()
    ):
        return True
    if isinstance(obj, list) and all(is_json(v) for v in obj):
        return True
    return False


@register_format
class JsonFormat(TextDiff, Format[object]):
    "Stores the data with `json.dump()`."

    suffix = ".json"
    priority = -10

    def is_format_for(self, value: object):
        return is_json(value)

    def encode(self, value: object, path: Path):

        with path.open("w", newline="\n", encoding="utf-8") as f:
            json.dump(value, f, indent=2, ensure_ascii=False)

    def decode(self, path: Path) -> object:

        with path.open("r", newline="\n", encoding="utf-8") as f:
            return json.load(f)



from __future__ import annotations

import dataclasses
import re
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext


class Location:
    suffix: str

    def __str__(self) -> str:
        raise NotImplementedError

    @contextmanager
    def load(self) -> Generator[Path]:
        raise NotImplementedError

    def store(self, new_file: Path):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError

    def exists(self):
        raise NotImplementedError


@dataclass
class ExternalLocation(Location):
    storage: str
    stem: str
    suffix: str

    filename: Path | None = None
    qualname: str | None = None
    linenumber: int | None = None

    @classmethod
    def from_name(
        cls,
        name: str | None,
        *,
        context: AdapterContext | None = None,
        filename: Path | None = None,
    ):
        from inline_snapshot._global_state import state

        if not name:
            storage = state().config.default_storage
            stem = ""
            suffix = ""
        else:
            m = re.fullmatch(r"([0-9a-fA-F]{64}|[0-9a-fA-F]+\*)(\.[a-zA-Z0-9]+)", name)

            if m:
                storage = "hash"
                path = name
            elif ":" in name:
                storage, path = name.split(":", 1)
                if storage not in ("hash", "uuid"):
                    raise ValueError(f"storage has to be hash or uuid")
            else:
                storage = state().config.default_storage
                path = name

            if "." in path:
                stem, suffix = path.split(".", 1)
                suffix = "." + suffix
            elif not path:
                stem = ""
                suffix = ""
            else:
                raise ValueError(f"'{name}' is missing a suffix")

        qualname = None
        if context:
            filename = Path(context.file.filename)
            qualname = context.qualname

        return cls(storage, stem, suffix, filename, qualname, None)

    @property
    def path(self) -> str:
        return f"{self.stem or ''}{self.suffix or ''}"

    def to_str(self):
        return str(self)

    def __str__(self) -> str:
        return f"{self.storage}:{self.path}"

    def with_stem(self, new_stem):
        return dataclasses.replace(self, stem=new_stem)

    @contextmanager
    def load(self) -> Generator[Path]:
        from inline_snapshot._global_state import state

        assert self.storage

        storage = state().all_storages[self.storage]
        with storage.load(self) as file:
            yield file

    def store(self, new_file: Path):
        from inline_snapshot._global_state import state

        assert self.storage

        storage = state().all_storages[self.storage]
        storage.store(self, new_file)

    def delete(self):
        from inline_snapshot._global_state import state

        assert self.storage

        storage = state().all_storages[self.storage]
        storage.delete(self)

    def exists(self):
        return self.stem


class FileLocation(Location):
    def __init__(self, filename: Path):
        self._filename = filename

    @property
    def suffix(self):
        return self._filename.suffix

    def __str__(self) -> str:
        p = self._filename.resolve()

        try:
            p = p.relative_to(Path.cwd().resolve())
        except ValueError:
            pass

        return p.as_posix()

    @contextmanager
    def load(self) -> Generator[Path]:
        yield self._filename

    def store(self, new_file: Path):
        self._filename.parent.mkdir(exist_ok=True, parents=True)
        shutil.copy(new_file, self._filename)

    def exists(self):
        return self._filename.exists()



from __future__ import annotations

import importlib.util
import os
import platform
import random
import re
import subprocess as sp
import sys
import traceback
import uuid
from argparse import ArgumentParser
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable
from unittest.mock import patch

from rich.console import Console

from inline_snapshot._exceptions import UsageError
from inline_snapshot._snapshot_session import SnapshotSession

from .._global_state import enter_snapshot_context
from .._global_state import leave_snapshot_context
from .._global_state import state
from .._types import Category
from .._types import Snapshot

ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)


def normalize(text):
    text = ansi_escape.sub("", text)

    # fix windows problems
    text = text.replace("\u2500", "-")
    text = text.replace("\r", "")
    return text


@contextmanager
def deterministic_uuid():
    rd = random.Random(0)

    def f():
        return uuid.UUID(int=rd.getrandbits(128), version=4)

    with patch("uuid.uuid4", new=f):
        yield


# this code is copied from pytest

# Regex to match the session duration string in the summary: "74.34s".
rex_session_duration = re.compile(r"\d+\.\d\ds")
# Regex to match all the counts and phrases in the summary line: "34 passed, 111 skipped".
rex_outcome = re.compile(r"(\d+) (\w+)")


def parse_outcomes(lines):
    for line in reversed(lines):
        if rex_session_duration.search(line):
            outcomes = rex_outcome.findall(line)
            ret = {noun: int(count) for (count, noun) in outcomes}
            break
        else:
            pass  # pragma: no cover
    else:
        raise ValueError("Pytest terminal summary report not found")  # pragma: no cover

    to_plural = {
        "warning": "warnings",
        "error": "errors",
    }
    return {to_plural.get(k, k): v for k, v in ret.items()}


conftest_footer = """
import uuid
import random

rd = random.Random(0)

def f():
    return uuid.UUID(int=rd.getrandbits(128), version=4)

uuid.uuid4 = f
"""


@contextmanager
def chdir(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def change_file(path: Path, map_function):
    exists = path.exists()
    if exists:
        text = path.read_bytes().decode("utf-8")
    else:
        text = ""

    path.write_bytes(map_function(text).encode("utf-8"))

    yield

    if exists:
        path.write_bytes(text.encode("utf-8"))
    else:
        path.unlink()


@contextmanager
def temp_environ(**kwargs):
    original = dict(os.environ)
    os.environ.update(kwargs)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(original)


class StopTesting(Exception):
    pass


class Example:
    files: dict[str, str | bytes]

    def __init__(self, files: str | bytes | dict[str, str | bytes]):
        """
        Parameters:
            files: a collection of files which are used as your example project,
                   or just a string which will be saved as *tests/test_something.py*.
        """
        if isinstance(files, (str, bytes)):
            files = {"tests/test_something.py": files}

        self.files = files

    def dump_files(self):
        from rich.panel import Panel
        from rich.text import Text

        console = Console()

        for name, content in self.files.items():
            if isinstance(content, bytes):
                content = repr(content)
            console.print(Panel(Text(content), title=name))

    def _write_files(self, dir: Path):
        for name, content in self.files.items():
            filename = dir / name
            filename.parent.mkdir(exist_ok=True, parents=True)
            if isinstance(content, str):
                content = content.encode("utf-8")
            filename.write_bytes(content)

    def _read_files(self, dir: Path):

        def try_read(path: Path) -> str | bytes:
            code = path.read_bytes()
            try:
                return code.decode("utf-8")
            except UnicodeDecodeError:
                return code

        def normalize_path(path):
            return str(path.relative_to(dir)).replace("\\", "/")

        return {
            normalize_path(p): try_read(p)
            for p in dir.rglob("*")
            if p.is_file()
            and p.name != ".gitignore"
            and p.suffix != ".pyc"
            and ".pytest_cache" not in p.parts
        }

    def with_files(self, extra_files: dict[str, str | bytes]) -> Example:
        """
        Adds extra files to the example.

        Arguments:
            extra_files: dictionary of filenames and file contents.
        """
        return Example({**self.files, **extra_files})

    def read_file(self, filename: str) -> str:
        """
        Reads a file from the example.

        Arguments:
            filename: the filename.
        """
        text = self.files[filename]
        assert isinstance(text, str)
        return text

    read_text = read_file

    def change_code(self, mapping: Callable[[str], str]) -> Example:
        """
        Changes example tests by mapping every file with the given function.

        Arguments:
            mapping: function to apply to each file's content.
        """
        return Example(
            {
                name: mapping(text) if isinstance(text, str) else text
                for name, text in self.files.items()
            }
        )

    def replace(self, old_text: str, new_text: str) -> Example:
        """
        Changes example tests by replacing old_text with new_text.

        Arguments:
            old_text: the text to be replaced.
            new_text: the new text to use instead.
        """
        return self.change_code(lambda code: code.replace(old_text, new_text))

    def remove_file(self, filename: str) -> Example:
        """
        Removes a file from the example.

        Arguments:
            filename: the file to be removed.
        """
        return Example(
            {name: file for name, file in self.files.items() if name != filename}
        )

    def run_inline(
        self,
        args: list[str] = [],
        *,
        reported_categories: Snapshot[list[Category]] | None = None,
        changed_files: Snapshot[dict[str, str]] | None = None,
        report: Snapshot[str] | None = None,
        raises: Snapshot[str] | None = None,
        stderr: Snapshot[str] | None = None,
    ) -> Example:
        """Execute the example files in process and run every `test_*`
        function.

        This is useful for fast test execution.

        Parameters:
            args: inline-snapshot arguments (supports only "--inline-snapshot=fix|create|...").
            reported_categories: snapshot of categories which inline-snapshot thinks could be applied.
            changed_files: snapshot of files which are changed by this run.
            raises: snapshot of the exception raised during test execution.
                    Required if your code raises an exception.

        Returns:
            A new Example instance containing the changed files.
        """

        parser = ArgumentParser()

        parser.add_argument(
            "--inline-snapshot",
            metavar="(disable,short-report,report,review,create,update,trim,fix)*",
            dest="inline_snapshot",
            help="update specific snapshot values:\n"
            "disable: disable the snapshot logic\n"
            "short-report: show a short report\n"
            "report: show a longer report with a diff report\n"
            "review: allow to approve the changes interactive\n"
            "create: creates snapshots which are currently not defined\n"
            "update: update snapshots even if they are defined\n"
            "trim: changes the snapshot in a way which will make the snapshot more precise.\n"
            "fix: change snapshots which currently break your tests\n",
        )
        parsed_args = parser.parse_args(args)
        flags = (parsed_args.inline_snapshot or "").split(",")
        self.dump_files()
        flags = {f for f in flags if f}

        with TemporaryDirectory() as dir:
            tmp_path = Path(dir)

            self._write_files(tmp_path)

            raised_exception = []

            with deterministic_uuid(), chdir(tmp_path), temp_environ(TERM="unknown"):
                session = SnapshotSession()

                def report_error(message):
                    raise StopTesting(message)

                snapshot_flags = set()
                old_modules = sys.modules
                old_path = sys.path[:]
                # Add tmp_path to sys.path so modules can be imported normally
                sys.path.insert(0, str(tmp_path))

                try:
                    enter_snapshot_context()
                    session.load_config(
                        tmp_path / "pyproject.toml",
                        flags,
                        parallel_run=False,
                        error=report_error,
                        project_root=tmp_path,
                    )

                    report_output = StringIO()
                    console = Console(file=report_output, width=80)

                    # Load and register all conftest.py files first
                    for conftest_path in tmp_path.rglob("conftest.py"):
                        print("load> conftest", conftest_path)

                        # Load conftest module using importlib
                        spec = importlib.util.spec_from_file_location(
                            f"conftest_{conftest_path.parent.name}", conftest_path
                        )
                        assert spec and spec.loader

                        conftest_module = importlib.util.module_from_spec(spec)
                        sys.modules[spec.name] = conftest_module
                        conftest_module.__file__ = str(conftest_path)
                        spec.loader.exec_module(conftest_module)

                        # Register customize hooks from this conftest
                        session.register_customize_hooks_from_module(conftest_module)

                    tests_found = False
                    for filename in tmp_path.rglob("test_*.py"):
                        print("run> pytest-inline", filename, *args)

                        # Load module using importlib
                        spec = importlib.util.spec_from_file_location(
                            filename.stem, filename
                        )
                        assert spec and spec.loader

                        module = importlib.util.module_from_spec(spec)
                        sys.modules[filename.stem] = module
                        spec.loader.exec_module(module)

                        # run all test_* functions
                        tests = [
                            v
                            for k, v in module.__dict__.items()
                            if (k.startswith("test_") or k == "test") and callable(v)
                        ]
                        tests_found |= len(tests) != 0

                        for v in tests:
                            try:

                                def fail(message):
                                    console.print(f"FAIL: {message}")

                                session.test_enter()
                                try:
                                    v()
                                finally:
                                    session.test_exit(fail=fail)
                            except Exception as e:
                                traceback.print_exc()
                                raised_exception.append(e)

                    if not tests_found:
                        raise UsageError("no test_*() functions in the example")

                    session.show_report(console)

                    for snapshot in state().snapshots.values():
                        for change in snapshot._changes():
                            snapshot_flags.add(change.flag)

                except StopTesting as e:
                    assert stderr == f"ERROR: {e}\n"
                finally:
                    sys.modules = old_modules
                    sys.path = old_path
                    leave_snapshot_context()

            if reported_categories is not None:
                assert sorted(snapshot_flags) == reported_categories

            if raised_exception:
                if raises is None:
                    raise raised_exception[0]

                assert raises == "\n".join(
                    f"{type(e).__name__}:\n" + str(e) for e in raised_exception
                )
            else:
                assert raises == None

            if changed_files is not None:
                assert changed_files == self._changed_files(tmp_path)

            if report is not None:
                assert report == normalize(report_output.getvalue())

            return Example(self._read_files(tmp_path))

    def _changed_files(self, tmp_path):
        current_files = {}
        all_current_files = self._read_files(tmp_path)

        for name, content in sorted(all_current_files.items()):
            if name not in self.files or self.files[name] != content:
                current_files[name] = content

        for name in self.files:
            if name not in all_current_files:
                current_files[name] = None

        return current_files

    def run_pytest(
        self,
        args: list[str] = [],
        *,
        term_columns=80,
        env: dict[str, str] = {},
        changed_files: Snapshot[dict[str, str]] | None = None,
        report: Snapshot[str] | None = None,
        error: Snapshot[str] | None = None,
        stderr: Snapshot[str] | None = None,
        returncode: Snapshot[int] = 0,
        stdin: bytes = b"",
        outcomes: Snapshot[dict[str, int]] | None = None,
    ) -> Example:
        """Run pytest with the given args and environment variables in a separate
        process.

        It can be used to test the interaction between your code and pytest, but it is a bit slower than `run_inline`.

        Parameters:
            args: pytest arguments like "--inline-snapshot=fix"
            env: dict of environment variables
            changed_files: snapshot of files changed by this run.
            report: snapshot of the report at the end of the pytest run.
            stderr: pytest stderr output
            returncode: snapshot of the pytest return code.

        Returns:
            A new Example instance containing the changed files.
        """

        self.dump_files()

        with TemporaryDirectory() as dir:
            tmp_path = Path(dir)

            self._write_files(tmp_path)

            cmd = [sys.executable, "-m", "pytest", *args]

            command_env = dict(os.environ)
            command_env["TERM"] = "unknown"
            command_env["COLUMNS"] = str(
                term_columns + 1 if platform.system() == "Windows" else term_columns
            )
            command_env.pop("CI", None)
            command_env.pop("GITHUB_ACTIONS", None)
            command_env.pop("PYTEST_XDIST_WORKER", None)

            if stdin:
                # makes Console.is_terminal == True
                command_env["FORCE_COLOR"] = "true"

            command_env.update(env)

            with change_file(
                tmp_path / "conftest.py", lambda text: text + conftest_footer
            ):
                result = sp.run(
                    cmd, cwd=tmp_path, capture_output=True, env=command_env, input=stdin
                )

            result_stdout = result.stdout.decode("utf-8")
            result_stderr = result.stderr.decode("utf-8")

            result_returncode = result.returncode

            print("run>", *cmd)
            print("stdout:")
            print(result_stdout)
            print("stderr:")
            print(result_stderr)

            assert result.returncode == returncode

            if stderr is not None:

                original = result_stderr.splitlines()
                lines = [
                    line
                    for line in original
                    if not any(
                        s in line
                        for s in [
                            'No entry for terminal type "unknown"',
                            "using dumb terminal settings.",
                        ]
                    )
                ]

                assert "\n".join(lines) == stderr

            if report is not None:

                report_list = []
                record = False
                for line in result_stdout.splitlines():
                    line = normalize(line.strip())
                    if line.startswith("===="):
                        record = False

                    if record and line:
                        report_list.append(line)

                    if (
                        line.startswith(("-----", "═════"))
                        and "inline-snapshot" in line
                    ):
                        record = True

                report_str = "\n".join(report_list)

                assert report_str == report, repr(report_str)

            if error is not None:
                assert (
                    error
                    == "\n".join(
                        [
                            line
                            for line in result_stdout.splitlines()
                            if line and line[:2] in ("> ", "E ")
                        ]
                    )
                    + "\n"
                )

            if changed_files is not None:
                assert changed_files == self._changed_files(tmp_path)

            if outcomes is not None:
                assert outcomes == parse_outcomes(result_stdout.splitlines())

            return Example(self._read_files(tmp_path))



from __future__ import annotations

import ast
from dataclasses import dataclass
from dataclasses import field
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase

from ._custom import Custom


class CustomSequenceTypes:
    trailing_comma: bool
    braces: str
    value_type: type


@dataclass(frozen=True)
class CustomSequence(Custom, CustomSequenceTypes):
    value: list[Custom] = field(compare=False)

    def _map(self, f):
        return f(self.value_type([x._map(f) for x in self.value]))

    def _code_repr(self, context: AdapterContext) -> Generator[ChangeBase, None, str]:
        values = []
        for v in self.value:
            value = yield from v._code_repr(context)
            values.append(value)

        trailing_comma = self.trailing_comma and len(self.value) == 1
        return f"{self.braces[0]}{', '.join(values)}{', ' if trailing_comma else ''}{self.braces[1]}"


class CustomList(CustomSequence):
    node_type = ast.List
    value_type = list
    braces = "[]"
    trailing_comma = False


class CustomTuple(CustomSequence):
    node_type = ast.Tuple
    value_type = tuple
    braces = "()"
    trailing_comma = True



"""The following functions are built on top of inline-snapshot and could also
be implemented in an extra library.
The source is added to the documentation, which allows you to look at how they are implemented and implement similar functions if you need them.

They are part of inline-snapshot because they are generally useful and do
not depend on other libraries.
"""

import contextlib
import io
import warnings
from contextlib import redirect_stderr
from contextlib import redirect_stdout
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple
from typing import Union

from inline_snapshot._code_repr import code_repr

from ._types import Snapshot
from ._unmanaged import declare_unmanaged


@contextlib.contextmanager
def raises(exception: Snapshot[str]):
    """Check that an exception is raised.

    Parameters:
        exception: Snapshot that is compared with `#!python f"{type}: {message}"` if an exception occurs, or `#!python "<no exception>"` if no exception is raised.

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        from inline_snapshot import snapshot
        from inline_snapshot.extra import raises


        def test_raises():
            with raises(snapshot()):
                1 / 0
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="6"
        from inline_snapshot import snapshot
        from inline_snapshot.extra import raises


        def test_raises():
            with raises(snapshot("ZeroDivisionError: division by zero")):
                1 / 0
        ```
    """

    try:
        yield
    except BaseException as ex:
        msg = str(ex)
        if "\n" in msg:
            assert f"{type(ex).__name__}:\n{ex}" == exception
        else:
            assert f"{type(ex).__name__}: {ex}" == exception
    else:
        assert "<no exception>" == exception


@contextlib.contextmanager
def prints(*, stdout: Snapshot[str] = "", stderr: Snapshot[str] = ""):
    """Uses `contextlib.redirect_stderr/stdout` to capture the output and
    compare it with the snapshots. `dirty_equals.IsStr` can be used to ignore
    the output if needed.

    Parameters:
        stdout: Snapshot that is compared to the recorded output.
        stderr: Snapshot that is compared to the recorded error output.

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        import sys
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints


        def test_prints():
            with prints(stdout=snapshot(), stderr=snapshot()):
                print("hello world")
                print("some error", file=sys.stderr)
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="7 8 9"
        import sys
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints


        def test_prints():
            with prints(
                stdout=snapshot("hello world\\n"), stderr=snapshot("some error\\n")
            ):
                print("hello world")
                print("some error", file=sys.stderr)
        ```

    === "ignore stdout"

        <!-- inline-snapshot: outcome-passed=1 -->
        ``` python hl_lines="2 9 10"
        import sys
        from dirty_equals import IsStr
        from inline_snapshot import snapshot
        from inline_snapshot.extra import prints


        def test_prints():
            with prints(
                stdout=IsStr(),
                stderr=snapshot("some error\\n"),
            ):
                print("hello world")
                print("some error", file=sys.stderr)
        ```
    """

    with redirect_stdout(io.StringIO()) as stdout_io:
        with redirect_stderr(io.StringIO()) as stderr_io:
            yield

    assert stderr_io.getvalue() == stderr
    assert stdout_io.getvalue() == stdout


Warning = Union[str, Tuple[int, str], Tuple[str, str], Tuple[str, int, str]]


@contextlib.contextmanager
def warns(
    expected_warnings: Snapshot[List[Warning]],
    /,
    include_line: bool = False,
    include_file: bool = False,
):
    """
    Captures warnings with `warnings.catch_warnings` and compares them against expected warnings.

    Parameters:
        expected_warnings: Snapshot containing a list of expected warnings.
        include_line: If `True`, each expected warning is a tuple `(line_number, message)`.
        include_file: If `True`, each expected warning is a tuple `(filename, message)`.

    The format of the expected warning:

    - `(filename, line_number, message)` if both `include_line` and `include_file` are `True`.
    - `(line_number, message)` if only `include_line` is `True`.
    - `(filename, message)` if only `include_file` is `True`.
    - A string `message` if both are `False`.

    === "original"

        <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
        ``` python
        from warnings import warn
        from inline_snapshot import snapshot
        from inline_snapshot.extra import warns


        def test_warns():
            with warns(snapshot(), include_line=True):
                warn("some problem")
        ```

    === "--inline-snapshot=create"

        <!-- inline-snapshot: create fix outcome-passed=1 outcome-errors=1 -->
        ``` python hl_lines="7"
        from warnings import warn
        from inline_snapshot import snapshot
        from inline_snapshot.extra import warns


        def test_warns():
            with warns(snapshot([(8, "UserWarning: some problem")]), include_line=True):
                warn("some problem")
        ```
    """
    with warnings.catch_warnings(record=True) as result:
        warnings.simplefilter("always")
        yield

    def make_warning(w):
        message = f"{w.category.__name__}: {w.message}"
        if not include_line and not include_file:
            return message
        message = (message,)

        if include_line:
            message = (w.lineno, *message)
        if include_file:
            message = (w.filename, *message)

        return message

    assert [make_warning(w) for w in result] == expected_warnings


@declare_unmanaged
class Transformed:
    """
    `Transformed` allows you to move transformations of your values from one side of the `==` to the other.

    <!-- inline-snapshot: first_block outcome-passed=1 outcome-errors=1 -->
    ``` python
    from inline_snapshot import snapshot
    from inline_snapshot.extra import Transformed


    def test_transform():
        numbers = [1, 8, 3, 7, 5]
        assert sorted(numbers) == snapshot()
        assert numbers == Transformed(sorted, snapshot())
    ```

    Both assertions create the same snapshots.

    <!-- inline-snapshot: create outcome-passed=1 outcome-errors=1 -->
    ``` python hl_lines="7 8"
    from inline_snapshot import snapshot
    from inline_snapshot.extra import Transformed


    def test_transform():
        numbers = [1, 8, 3, 7, 5]
        assert sorted(numbers) == snapshot([1, 3, 5, 7, 8])
        assert numbers == Transformed(sorted, snapshot([1, 3, 5, 7, 8]))
    ```

    `Transformed` is more flexible to use because you can also use it deep inside data structures.
    The following example shows how `Transformed` is used inside a dictionary.

    <!-- inline-snapshot: create fix first_block outcome-passed=1 -->
    ``` python
    from random import shuffle
    from inline_snapshot import snapshot
    from inline_snapshot.extra import Transformed


    def request():
        data = [1, 8, 18748, 493]
        shuffle(data)
        return {"name": "example", "data": data}


    def test_request():
        assert request() == snapshot(
            {
                "name": "example",
                "data": Transformed(sorted, snapshot([1, 8, 493, 18748])),
            }
        )
    ```

    Or to normalize strings.

    <!-- inline-snapshot: create fix first_block outcome-passed=1 -->
    ``` python
    import re
    from inline_snapshot import snapshot
    from inline_snapshot.extra import Transformed


    class Thing:
        def __repr__(self):
            return "<Thing with some random id 152897513>"


    def without_ids(text):
        return re.sub(r"<([^0-9]*)[^>]+>", lambda m: f"<{m[1]} ...>", text)


    def test_text_with_objects():
        text = f"text can contain {Thing()}"

        assert {"logs": text} == snapshot(
            {
                "logs": Transformed(
                    without_ids,
                    snapshot("text can contain <Thing with some random id  ...>"),
                )
            }
        )
    ```


    !!! Tip
        You can use [@transformation][inline_snapshot.extra.transformation] if you want to use the same transformation multiple times.

    """

    def __init__(
        self, func: Callable[[Any], Any], value: Snapshot, should_be: Any = None
    ) -> None:
        """
        Arguments:
            func: functions which is used to transform the value which is compared.
            value: the result of the transformation
            should_be: this argument is unused and only reported in the `repr()` to show you the last transformed value.
        """
        self._func = func
        self._value = value
        self._last_transformed_value = None

    def __eq__(self, other) -> bool:
        self._last_transformed_value = self._func(other)
        return self._last_transformed_value == self._value

    def __repr__(self):
        if self._last_transformed_value == self._value:
            return f"Transformed({code_repr(self._func)}, {self._value})"
        else:
            return f"Transformed({code_repr(self._func)}, {self._value}, should_be={self._last_transformed_value!r})"


def transformation(func):
    """

    `@transformation` can be used to bind a function to [Transformed][inline_snapshot.extra.Transformed],
    which simplifies your code if you want to use the same transformation multiple times.

    <!-- inline-snapshot: create first_block outcome-passed=1 -->
    ``` python
    import re
    from inline_snapshot import snapshot
    from inline_snapshot.extra import transformation


    class Thing:
        def __repr__(self):
            return "<Thing with some random id 152897513>"


    @transformation
    def WithoutIds(text):
        return re.sub(r"<([^0-9]*)[^>]+>", lambda m: f"<{m[1]} ...>", text)


    def test_text_with_objects():
        text = f"text can contain {Thing()}"

        assert {"logs": [text]} == snapshot(
            {
                "logs": [
                    WithoutIds(
                        snapshot(
                            "text can contain <Thing with some random id  ...>"
                        )
                    )
                ]
            }
        )
    ```

    !!! Tip
        The argument of `WithoutIds` can also be an external `WithoutIds(external())` if you want to store a large log in an external file.
    """

    def f(value):
        return Transformed(func, value)

    return f



from __future__ import annotations

from pathlib import Path

from inline_snapshot._change import ExternalChange
from inline_snapshot._exceptions import UsageError
from inline_snapshot._external._format._protocol import get_format_handler
from inline_snapshot._external._outsource import Outsourced
from inline_snapshot._external._storage._protocol import StorageLookupError
from inline_snapshot._global_state import state

from .._snapshot.generic_value import GenericValue
from ._external_location import ExternalLocation
from ._external_location import Location


class ExternalBase:

    _original_location: Location
    _location: Location
    _tmp_file: Path | None

    def __init__(self):
        self._tmp_file = None
        self._was_compared = False

    def _changes(self):
        if self._tmp_file is not None:
            yield ExternalChange(
                "create" if self._is_empty() else "fix",
                self._tmp_file,
                self._original_location,
                self._location,
                self._format,
            )

    def _assign(self, value):
        raise NotImplementedError()

    def _load_value(self):
        raise NotImplementedError()

    def __eq__(self, other):
        """Two external objects are equal if they have the same value"""

        external_type = (
            "external" if type(self).__name__ == "External" else "external_file"
        )
        __tracebackhide__ = True

        if isinstance(other, Outsourced):

            format = get_format_handler(other.data, other.suffix or "")

            self._location.suffix = other.suffix or format.suffix
            other = other.data

        elif isinstance(other, ExternalBase):
            raise UsageError(
                f"you can not compare {external_type}(...) with {external_type}(...)"
            )

        elif isinstance(other, GenericValue):
            raise UsageError(
                f"you can not compare {external_type}(...) with snapshot(...)"
            )

        first_comparison = not self._was_compared

        self._was_compared = True

        if self._is_empty():
            self._assign(other)
            state().missing_values += 1
            if state().update_flags.create:
                return True
            return False

        try:
            value = self._load_value()
        except StorageLookupError as error:
            if not error.files and state().update_flags.fix:
                self._original_location = ExternalLocation.from_name("")
                self._assign(other)
                state().incorrect_values += 1
                return True
            else:
                raise

        result = value == other

        if not result and first_comparison:
            self._assign(other)
            state().incorrect_values += 1
            return True
        return result



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Protocol
from typing import TypeVar

from rich.console import RenderableType

from inline_snapshot._exceptions import UsageError


def get_format_handler(data, suffix: str) -> Format:
    """
    Retrieves the appropriate format handler for the given data and suffix.

    Arguments:
        data: The data to be formatted.
        suffix: The suffix associated with the desired format.

    Returns:
        The format handler that matches the data and suffix.

    Raises:
        UsageError: If no format handler is found or if multiple handlers
        with the same priority exist for the given data and suffix.
    """
    from inline_snapshot._global_state import state

    if suffix:
        suffix = state().format_aliases.get(suffix, suffix)

    possible_formatter = sorted(
        [
            formatter
            for formatter in state().all_formats.values()
            if formatter.is_format_for(data)
            and (not suffix or suffix == formatter.suffix)
        ],
        key=lambda format: format.priority,
    )

    if not possible_formatter:
        and_suffix = f" and suffix '{suffix}'" if suffix else ""
        raise UsageError(
            f"No format handler found for the given type '{type(data).__qualname__}'{and_suffix}."
        )
    if (
        len(possible_formatter) >= 2
        and possible_formatter[-1].priority == possible_formatter[-2].priority
    ):
        and_suffix = f" and suffix '{suffix}'" if suffix else ""
        raise UsageError(
            f"Multiple format handlers found for the given type '{type(data).__qualname__}'. "
            f"The following handlers have the same priority: {[f.suffix for f in possible_formatter]}. "
            f"You can explicitly choose one with external('.suffix') or adjust the priorities of the handlers if you implemented them."
        )

    return possible_formatter[-1]


def get_format_handler_from_suffix(suffix: str) -> Format:
    """
    Retrieves the format handler associated with the given suffix.

    Arguments:
        suffix: The suffix for which the format handler is required.

    Returns:
        The format handler corresponding to the suffix.

    Raises:
        UsageError: If no format handler is found for the given suffix.
    """
    from inline_snapshot._global_state import state

    suffix = state().format_aliases.get(suffix, suffix)

    format = state().all_formats.get(suffix, None)

    if format is None:
        raise UsageError(f"Format '{suffix}' is unknown.")

    return format


T = TypeVar("T")


class Format(Protocol[T]):
    """
    Base class for the Format Protocol.
    """

    suffix: str
    """
    The suffix associated with this format handler.

    Every format implementation must define a suffix. This suffix is used
    when the external file is written and is required to find the correct
    format handler when the file is read again.
    """

    priority: int = 0
    """
    Determines the correct format when multiple format handlers can handle
    a given value (`is_format_for`).

    *priority* is `0` by default and can be set to a smaller number for
    generic formats that work with many data types (e.g., *.json*), where
    `is_format_for()` also returns `True` for `str` and `bytes`. This allows
    you to use `assert "text" == external()` without explicitly providing
    a `".txt"` suffix to distinguish it from `".json"`.

    A higher *priority* can be used for more specific formats, such as bytes
    with a `b"\\x89PNG"` prefix that should be stored as *.png* files.
    """

    def rich_diff(self, original: Path, new: Path) -> RenderableType:
        """
        Displays a diff between the original and new files.

        Arguments:
            original: The path to the original external file.
            new: The path to the new external file.

        Returns:
            A rich renderable object representing the diff. This can be a
            textual diff or another type of representation.
        """
        raise NotImplementedError

    def rich_show(self, path: Path) -> RenderableType:
        """
        Displays a representation of a newly created external object.

        Arguments:
            path: The path to the external file.

        Returns:
            A rich renderable object representing the new object. The
            representation should be concise.
        """
        raise NotImplementedError

    def is_format_for(self, value: object) -> bool:
        """
        Determines if this format handler can handle the given value.

        This function is used to find the correct format handler when no
        suffix is provided.

        ``` python
        assert value == external()
        ```

        Arguments:
            value: The value to be formatted.

        Returns:
            `True` if the value is handled by this format implementation.
        """
        raise NotImplementedError

    def encode(self, value: T, path: Path) -> None:
        """
        Converts the value and stores it in the specified path.

        Arguments:
            value: The value to be stored.
            path: The path to a temporary file where the value should be stored.
        """
        raise NotImplementedError

    def decode(self, path: Path) -> T:
        """
        Reads the value from the specified path.

        Arguments:
            path: The path to a temporary file where the value is stored.

        Returns:
            The value of the external object.
        """
        raise NotImplementedError


FormatT = TypeVar("FormatT")


def register_format(
    format: type[Format[FormatT]] | Format[FormatT] | None = None,
    *,
    replace_handler: bool = False,
):
    """
    Registers a new format handler for the suffix `format.suffix`.

    This function can also be used as a decorator:

    ``` python
    @register_format
    class MyFormat: ...
    ```

    which is equivalent to:

    ``` python
    register_format(MyFormat())
    ```

    Arguments:
        format: The format handler class or instance.
        replace_handler: If `True`, replaces an existing handler for the same suffix.

    Raises:
        UsageError: If a handler for the same suffix already exists and `replace_handler` is not set to `True`.
    """
    if format is None:
        return partial(register_format, replace_handler=replace_handler)

    from inline_snapshot._global_state import state

    instance = format() if isinstance(format, type) else format

    if not replace_handler and instance.suffix in state().all_formats:
        raise UsageError(
            f"A format handler is already registered for the suffix '{instance.suffix}'."
        )

    state().all_formats[instance.suffix] = instance
    return format


def register_format_alias(suffix: str, format: str):
    """
    Registers an alias for a given format suffix.

    Arguments:
        suffix: The suffix to register the alias for.
        format: The suffix of the format that should be used instead.

    Notes:
        The alias suffix is used to find the
        correct format handler.
    """
    from inline_snapshot._global_state import state

    state().format_aliases[suffix] = format



from __future__ import annotations

import contextlib
from copy import deepcopy
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING
from typing import Any
from typing import Generator
from typing import Literal
from uuid import uuid4

import pluggy

from inline_snapshot._config import Config
from inline_snapshot.plugin._spec import inline_snapshot_plugin_name

if TYPE_CHECKING:
    from inline_snapshot._external._format._protocol import Format
    from inline_snapshot._external._storage._protocol import StorageProtocol
    from inline_snapshot._types import SnapshotRefBase

from ._flags import Flags


@dataclass
class State:
    config: Config = field(default_factory=Config)

    # snapshot
    missing_values: int = 0
    incorrect_values: int = 0

    snapshots: dict[Any, SnapshotRefBase] = field(default_factory=dict)
    update_flags: Flags = field(default_factory=Flags)
    active: bool = True

    all_problems: set[str] = field(default_factory=set)

    flags: set[str] = field(default_factory=set)

    format_aliases: dict[str, str] = field(default_factory=dict)

    all_formats: dict[str, Format] = field(default_factory=dict)

    all_storages: dict[str, StorageProtocol] = field(default_factory=dict)

    default_storage: str = "uuid"

    tmp_dir: TemporaryDirectory | None = field(
        default_factory=lambda: TemporaryDirectory(prefix="inline-snapshot-")
    )

    pm: pluggy.PluginManager = field(
        default_factory=lambda: pluggy.PluginManager(inline_snapshot_plugin_name)
    )

    def new_tmp_path(self, suffix: str) -> Path:
        assert self.tmp_dir is not None
        return Path(self.tmp_dir.name) / f"tmp-path-{uuid4()}{suffix}"

    disable_reason: Literal["xdist", "ci", "implementation", None] = None


_latest_global_states: list[State] = []

_current: State = State(active=False, tmp_dir=None)


def state() -> State:
    global _current
    return _current


def enter_snapshot_context():
    global _current
    latest = _current
    _latest_global_states.append(_current)
    _current = State()
    _current.all_formats = dict(latest.all_formats)
    _current.config = deepcopy(latest.config)

    from .plugin._spec import InlineSnapshotPluginSpec

    _current.pm.add_hookspecs(InlineSnapshotPluginSpec)

    from .plugin._default_plugin import InlineSnapshotPlugin

    _current.pm.register(InlineSnapshotPlugin())

    try:
        from .plugin._default_plugin import InlineSnapshotAttrsPlugin
    except ImportError:  # pragma: no cover
        pass
    else:
        _current.pm.register(InlineSnapshotAttrsPlugin())

    try:
        from .plugin._default_plugin import InlineSnapshotPydanticPlugin
    except ImportError:  # pragma: no cover
        pass
    else:
        _current.pm.register(InlineSnapshotPydanticPlugin())

    try:
        from .plugin._default_plugin import InlineSnapshotDirtyEqualsPlugin
    except ImportError:  # pragma: no cover
        pass
    else:
        _current.pm.register(InlineSnapshotDirtyEqualsPlugin())

    _current.pm.load_setuptools_entrypoints(inline_snapshot_plugin_name)


def leave_snapshot_context():
    global _current
    _current.tmp_dir.cleanup()
    _current = _latest_global_states.pop()


@contextlib.contextmanager
def snapshot_env() -> Generator[State]:
    from ._external._storage._hash import HashStorage

    old = _current

    enter_snapshot_context()

    try:
        with TemporaryDirectory() as dir:
            _current.all_storages = dict(old.all_storages)
            _current.all_storages["hash"] = HashStorage(dir)

            yield _current
    finally:
        leave_snapshot_context()


def state_cached(function):
    def w():
        name = "_cached_" + function.__name__
        if not hasattr(state(), name):
            setattr(state(), name, function())

        return getattr(state(), name)

    return w



from __future__ import annotations

import ast
from pathlib import Path

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import CallArg
from inline_snapshot._change import Replace
from inline_snapshot._exceptions import UsageError
from inline_snapshot._external._external_base import ExternalBase
from inline_snapshot._external._format._protocol import get_format_handler
from inline_snapshot._external._format._protocol import get_format_handler_from_suffix
from inline_snapshot._global_state import state
from inline_snapshot._inline_snapshot import create_snapshot
from inline_snapshot._unmanaged import declare_unmanaged
from inline_snapshot._utils import is_relative_to

from ._external_location import ExternalLocation


def external(name: str | None = None):
    return create_snapshot(External, name)


def is_inside_testdir(path: Path) -> bool:
    path = path.resolve()
    for dir in state().config.test_directories or []:
        if is_relative_to(dir.resolve(), path):
            return True
    return False


@declare_unmanaged
class External(ExternalBase):
    _original_location: ExternalLocation
    _location: ExternalLocation

    def __init__(self, name: str, expr, context: AdapterContext):
        """External objects are used to move some data outside the source code.
        You should not instantiate this class directly, but by using `external()` instead.

        The location where the external data is stored can be defined by the storage protocol.
        The format is defined by the suffix.

        Parameters:
            name: the name of the external stored object which has the form "<hash>:<name>.<suffix>".
        """

        self._expr = expr
        self._context = context
        self._original_name = name

        self._location = ExternalLocation.from_name(name, context=context)
        self._original_location = ExternalLocation.from_name(name, context=context)

        super().__init__()

    @staticmethod
    def check_context(context: AdapterContext):
        if not is_inside_testdir(Path(context.file.filename)):
            raise UsageError(
                "external() can only be used in files which are inside tests/ or any other folder defined by your tool.inline-snapshot.test-dir in pyproject.toml"
            )

    def result(self):
        return self

    @classmethod
    def create_raw(cls, obj, context: AdapterContext):
        return cls._load_value_from_location(
            ExternalLocation.from_name(obj, context=context)
        )

    @property
    def _format(self):
        return get_format_handler_from_suffix(self._location.suffix or "")

    def _changes(self):
        if self._expr is None:
            node = None
        else:
            node = self._expr.node
            assert isinstance(node, ast.Call)

        new_name = self._location.to_str()
        if new_name != self._original_name:
            if self._original_name is None:
                yield CallArg(
                    "create",
                    self._context.file,
                    node,
                    0,
                    None,
                    f'"{new_name}"',
                    new_name,
                )

            else:
                yield Replace(
                    (
                        ("fix" if self._original_location.stem else "create")
                        if self._tmp_file is not None
                        else "update"
                    ),
                    self._context.file,
                    node.args[0] if node else None,
                    f'"{new_name}"',
                    self._original_name,
                    new_name,
                )

        yield from super()._changes()

    def _is_empty(self):
        return not self._original_location.stem

    def __repr__(self):
        """Returns the representation of the external object.

        The length of the hash can be specified in the
        [config](configuration.md).
        """

        return f'external("{self._original_location.to_str()}")'

    @property
    def storage(self):
        storage_name = self._location.storage or state().default_storage

        return state().all_storages[storage_name]

    def _assign(self, other):
        format = get_format_handler(other, self._location.suffix)

        if not self._location.suffix:
            self._location.suffix = format.suffix

        self._tmp_file = state().new_tmp_path(self._location.suffix)

        format.encode(other, self._tmp_file)
        self._location = self.storage.new_location(self._location, self._tmp_file)

    def _load_value(self):
        return self._load_value_from_location(self._original_location)

    @classmethod
    def _load_value_from_location(cls, location: ExternalLocation) -> object:
        assert location.storage

        if not location.stem:
            raise UsageError(
                f"can not load external object from an non existing location {location.to_str()!r}"
            )
        storage = state().all_storages[location.storage]

        with storage.load(location) as f:
            assert location.suffix
            format = get_format_handler_from_suffix(location.suffix)

            return format.decode(f)



import ast
from typing import Generator
from typing import Iterator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._code_repr import mock_repr
from inline_snapshot._customize._builder import Builder
from inline_snapshot._customize._custom import Custom
from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._new_adapter import reeval

from .._change import ChangeBase
from .._global_state import state
from .._types import SnapshotBase
from .._unmanaged import declare_unmanaged


def ignore_old_value():
    return state().update_flags.fix or state().update_flags.update


@declare_unmanaged
class GenericValue(SnapshotBase):
    _new_value: Custom
    _old_value: Custom
    _current_op = "undefined"
    _ast_node: ast.expr
    _context: AdapterContext

    def get_builder(self, **args):
        return Builder(_snapshot_context=self._context, **args)

    def _return(self, result, new_result=True):

        if not result:
            state().incorrect_values += 1
        flags = state().update_flags

        if (
            flags.fix
            or flags.create
            or flags.update
            or isinstance(self._old_value, CustomUndefined)
        ):
            return new_result
        return result

    @property
    def _file(self):
        return self._context.file

    def to_custom(self, value, **args):
        with mock_repr(self._context):
            return self.get_builder(**args)._get_handler(value)

    def value_to_custom(self, value):
        if isinstance(value, Custom):
            return value

        if self._ast_node is None:
            from inline_snapshot._snapshot.undecided_value import ValueToCustom

            return ValueToCustom(self._context).convert(value)
        else:
            from inline_snapshot._snapshot.undecided_value import AstToCustom

            return AstToCustom(self._context).convert(value, self._ast_node)

    def _re_eval(self, value, context: AdapterContext):
        self._context = context

        self._old_value = reeval(self._old_value, self.value_to_custom(value))

    def _ignore_old(self):
        return (
            state().update_flags.fix
            or state().update_flags.update
            or state().update_flags.create
            or isinstance(self._old_value, CustomUndefined)
        )

    def _visible_value(self):
        if self._ignore_old():
            return self._new_value
        else:
            return self._old_value

    def _get_changes(self) -> Iterator[ChangeBase]:
        raise NotImplementedError()

    def _new_code(self) -> Generator[ChangeBase, None, str]:
        raise NotImplementedError()

    def __repr__(self):
        return repr(self._visible_value()._eval())

    def _type_error(self, op):
        __tracebackhide__ = True
        raise TypeError(
            f"This snapshot cannot be use with `{op}`, because it was previously used with `{self._current_op}`"
        )

    def __eq__(self, other):
        __tracebackhide__ = True
        self._type_error("==")

    def __le__(self, other):
        __tracebackhide__ = True
        self._type_error("<=")

    def __ge__(self, other):
        __tracebackhide__ = True
        self._type_error(">=")

    def __contains__(self, item):
        __tracebackhide__ = True
        self._type_error("in")

    def __getitem__(self, item):
        __tracebackhide__ = True
        self._type_error("snapshot[key]")



from __future__ import annotations

import shutil
import uuid
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from inline_snapshot._global_state import state_cached
from inline_snapshot._problems import raise_problem
from inline_snapshot._utils import link

from .._external_location import ExternalLocation
from ._protocol import StorageLookupError
from ._protocol import StorageProtocol


@state_cached
def external_files() -> dict[str, Path]:
    from inline_snapshot._global_state import state

    base_folders = set()

    for test_dir in state().config.test_directories or []:
        base_folders |= set(test_dir.rglob("__inline_snapshot__"))

    return {
        file.name: file
        for folder in base_folders
        for file in folder.rglob("????????-????-????-????-????????????.*")
    }


class UuidStorage(StorageProtocol):
    @contextmanager
    def load(self, location: ExternalLocation) -> Generator[Path]:
        snapshot_path = self._lookup_path(location)

        yield snapshot_path

    def _lookup_path(self, location: ExternalLocation):
        if location.filename and location.qualname:
            path = self._get_path(location)
            if path.exists():
                return path

        if location.path in external_files():
            return external_files()[location.path]
        else:
            raise StorageLookupError(location, files=[])

    def store(self, location: ExternalLocation, file_path: Path):
        snapshot_path = self._get_path(location)

        snapshot_path.parent.mkdir(parents=True, exist_ok=True)

        shutil.copy(str(file_path), str(snapshot_path))

    def delete(self, location: ExternalLocation):
        snapshot_path = self._lookup_path(location)
        snapshot_path.unlink()

    def new_location(
        self, location: ExternalLocation, file_path: Path
    ) -> ExternalLocation:

        if not location.stem:
            return location.with_stem(str(uuid.uuid4()))
        return location

    def _get_path(self, location: ExternalLocation) -> Path:

        file_path = location.filename
        qualname = location.qualname
        assert file_path
        assert qualname

        if qualname == "<module>":
            qualname = "__module__"

        assert location.stem

        return (
            file_path.parent
            / "__inline_snapshot__"
            / f"{file_path.stem}"
            / qualname
            / f"{location.stem}{location.suffix}"
        )

    def find_unused_externals(
        self, used_externals: list[ExternalLocation]
    ) -> list[ExternalLocation]:

        used_names = [location.path for location in used_externals]

        unused_externals = {
            f.name for f in external_files().values() if f.name not in used_names
        }

        return [ExternalLocation.from_name("uuid:" + name) for name in unused_externals]

    def check_externals(self, used_externals: list[ExternalLocation]):
        grouped = defaultdict(list)
        for external in used_externals:
            grouped[external.to_str()].append(external)

        for external, externals in grouped.items():
            if len(externals) > 1:
                raise_problem(
                    f"The external {external} is used multiple times, which is not supported:\n"
                    + "\n".join(
                        sorted(
                            f"   {e.filename.resolve().relative_to(Path.cwd().resolve()).as_posix()}:{e.linenumber}"
                            for e in externals
                            if e.filename
                        )
                    )
                    + f"\n   (see {link('https://15r10nk.github.io/inline-snapshot/latest/external/external/#uuid')})"
                )



from pathlib import Path

from ._hash import HashStorage
from ._protocol import StorageLookupError
from ._protocol import StorageProtocol
from ._uuid import UuidStorage

__all__ = ("StorageLookupError", "StorageProtocol", "HashStorage")


def default_storages(storage_dir: Path):
    return {"hash": HashStorage(storage_dir / "external"), "uuid": UuidStorage()}



from typing import Generator
from typing import Iterator
from typing import List

from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._generator_utils import split_gen
from inline_snapshot._new_adapter import NewAdapter

from .._change import Change
from .._change import ChangeBase
from .._compare_context import compare_only
from .._global_state import state
from .generic_value import GenericValue


class EqValue(GenericValue):
    _current_op = "x == snapshot"
    _changes: List[Change]

    def __eq__(self, other):
        custom_other = self.to_custom(other, _build_new_value=True)

        if isinstance(self._old_value, CustomUndefined):
            state().missing_values += 1

        if not compare_only() and isinstance(self._new_value, CustomUndefined):
            self._changes = []

            adapter = NewAdapter(self._context)

            result = split_gen(
                adapter.compare(self._old_value, self._ast_node, custom_other)
            )
            self._changes = result.list
            self._new_value = result.value

        return self._return(
            self._old_value._eval() == other,
            self._new_value._eval() == other,
        )

    def _new_code(self) -> Generator[ChangeBase, None, str]:
        code = yield from self._new_value._code_repr(self._context)
        return code

    def _get_changes(self) -> Iterator[Change]:
        return iter(getattr(self, "_changes", []))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from inline_snapshot._external._format._protocol import get_format_handler
from inline_snapshot._global_state import state
from inline_snapshot._snapshot.generic_value import GenericValue


@dataclass
class Outsourced:
    data: Any
    suffix: str | None
    storage: str | None

    def __eq__(self, other):
        if isinstance(other, GenericValue):
            return NotImplemented

        if isinstance(other, Outsourced):
            return self.data == other.data

        from inline_snapshot._external._external_base import ExternalBase

        if isinstance(other, ExternalBase):
            return NotImplemented

        return self.data == other


def outsource(data: Any, suffix: str | None = None, storage: str | None = None) -> Any:
    if suffix and suffix[0] != ".":
        raise ValueError("suffix has to start with a '.' like '.png'")

    if not state().active:
        return data

    # check if the suffix/datatype is supported
    get_format_handler(data, suffix or "")

    return Outsourced(data, suffix, storage)



from typing import TypeVar

from ._exceptions import UsageError
from ._external._external import External
from ._external._external_file import ExternalFile
from ._external._outsource import Outsourced
from ._external._storage._protocol import StorageLookupError
from ._global_state import state
from ._is import Is
from ._snapshot.generic_value import GenericValue
from ._types import Snapshot


def unwrap(value):
    if isinstance(value, GenericValue):
        return value._visible_value()._map(lambda v: unwrap(v)[0]), True

    if isinstance(value, Outsourced):
        return (value.data, True)

    if isinstance(value, (External, Outsourced, ExternalFile)):
        try:
            return unwrap(value._load_value())[0], True
        except (UsageError, StorageLookupError):
            return (None, False)

    if isinstance(value, Is):
        return value.value, True

    return value, False


T = TypeVar("T", covariant=True)


def get_snapshot_value(snapshot_value: Snapshot[T]) -> T:
    """
    Extracts and returns the raw value stored inside a snapshot, removing all inline-snapshot-specific
    wrappers such as those generated by `external()`, `Is()`, or `snapshot()`.

    This function is primarily intended for extension authors who need direct access to the value
    of a previously stored snapshot. For standard test assertions, prefer using the snapshot directly.

    Args:
        snapshot_value: The snapshot object from which to extract the value.

    Returns:
        The unwrapped value contained in the snapshot.

    Example:
        ``` python
        from inline_snapshot import external, snapshot, get_snapshot_value

        s = snapshot([0, external("uuid:e3e70682-c209-4cac-a29f-6fbed82c07cd.json")])

        if record:
            # Store value
            assert s == [0, 5]
        else:
            # Use value from snapshot
            value = get_snapshot_value(s)
            # ... do something with value
        ```

    <!-- TODO: Find a way to test this code in the docs -->
    """
    if state().active:
        return unwrap(snapshot_value)[0]
    else:
        return snapshot_value  # type: ignore



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Callable

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._sentinels import undefined

missing = undefined
from inline_snapshot._compare_context import compare_context
from inline_snapshot._exceptions import UsageError

from ._custom import Custom
from ._custom_call import CustomCall
from ._custom_call import CustomDefault
from ._custom_code import CustomCode
from ._custom_code import Import
from ._custom_code import ImportFrom
from ._custom_dict import CustomDict
from ._custom_external import CustomExternal
from ._custom_sequence import CustomList
from ._custom_sequence import CustomTuple


@dataclass
class Builder:
    _snapshot_context: AdapterContext
    _build_new_value: bool = False
    _recursive: bool = True

    def _get_handler_recursive(self, v) -> Custom:
        if self._recursive:
            return self._get_handler(v)
        else:
            return v

    def _get_handler(self, v) -> Custom:

        from inline_snapshot._global_state import state

        result = v

        while not isinstance(result, Custom):
            with compare_context():
                r = state().pm.hook.customize(
                    value=result,
                    builder=self,
                    local_vars=self._local_vars,
                    global_vars=self._global_vars,
                )
            if r is None:
                result = CustomCode(result)
            else:
                result = r

        result.__dict__["original_value"] = v._eval() if isinstance(v, Custom) else v

        if not isinstance(v, Custom) and self._build_new_value:
            is_same = False
            v_eval = result._eval()

            if (
                hasattr(v, "__pydantic_generic_metadata__")
                and v.__pydantic_generic_metadata__["origin"] == v_eval
            ):
                is_same = True

            if not is_same and v_eval == v:
                is_same = True

            if not is_same:
                raise UsageError(
                    f"""\
Customized value does not match original value:

original_value={v!r}

customized_value={result._eval()!r}
customized_representation={result!r}
"""
                )

        return result

    def create_external(
        self, value: Any, format: str | None = None, storage: str | None = None
    ):
        """
        Creates a new `external()` with the given format and storage.
        """

        return CustomExternal(value, format=format, storage=storage)

    def create_list(self, value: list) -> Custom:
        """
        Creates an intermediate node for a list-expression which can be used as a result for your customization function.

        `create_list([1,2,3])` becomes `[1,2,3]` in the code.
        List elements don't have to be Custom nodes and are converted by inline-snapshot if needed.
        """
        custom = [self._get_handler_recursive(v) for v in value]
        return CustomList(value=custom)

    def create_tuple(self, value: tuple) -> Custom:
        """
        Creates an intermediate node for a tuple-expression which can be used as a result for your customization function.

        `create_tuple((1, 2, 3))` becomes `(1, 2, 3)` in the code.
        Tuple elements don't have to be Custom nodes and are converted by inline-snapshot if needed.
        """
        custom = [self._get_handler_recursive(v) for v in value]
        return CustomTuple(value=custom)

    def with_default(self, value: Any, default: Any):
        """
        Creates an intermediate node for a default value which can be used as an argument for create_call.

        Arguments are not included in the generated code when they match the actual default.
        The value doesn't have to be a Custom node and is converted by inline-snapshot if needed.
        """
        if isinstance(default, Custom):
            raise UsageError("default value can not be an Custom value")

        if value == default:
            return CustomDefault(value=self._get_handler_recursive(value))
        return value

    def create_call(
        self, function: Custom | Callable, posonly_args=[], kwargs={}
    ) -> Custom:
        """
        Creates an intermediate node for a function call expression which can be used as a result for your customization function.

        `create_call(MyClass, [arg1, arg2], {'key': value})` becomes `MyClass(arg1, arg2, key=value)` in the code.
        Function, arguments, and keyword arguments don't have to be Custom nodes and are converted by inline-snapshot if needed.
        """
        function = self._get_handler_recursive(function)
        posonly_args = [self._get_handler_recursive(arg) for arg in posonly_args]
        kwargs = {k: self._get_handler_recursive(arg) for k, arg in kwargs.items()}

        return CustomCall(
            function=function,
            args=posonly_args,
            kwargs=kwargs,
        )

    def create_dict(self, value: dict) -> Custom:
        """
        Creates an intermediate node for a dict-expression which can be used as a result for your customization function.

        `create_dict({'key': 'value'})` becomes `{'key': 'value'}` in the code.
        Dict keys and values don't have to be Custom nodes and are converted by inline-snapshot if needed.
        """
        custom = {
            self._get_handler_recursive(k): self._get_handler_recursive(v)
            for k, v in value.items()
        }
        return CustomDict(value=custom)

    @property
    def _local_vars(self):
        """Get local vars from snapshot context."""
        return self._snapshot_context.local_vars

    @property
    def _global_vars(self):
        """Get global vars from snapshot context."""
        return self._snapshot_context.global_vars

    def _build_import_vars(self, imports):
        """Build import vars from imports parameter."""
        import_vars = {}
        if imports:
            import importlib

            for imp in imports:
                if isinstance(imp, Import):
                    # import module - makes top-level package available
                    importlib.import_module(imp.module)
                    top_level = imp.module.split(".")[0]
                    import_vars[top_level] = importlib.import_module(top_level)
                elif isinstance(imp, ImportFrom):
                    # from module import name
                    module = importlib.import_module(imp.module)
                    import_vars[imp.name] = getattr(module, imp.name)
                else:
                    assert False
        return import_vars

    def create_code(
        self, code: str, *, imports: list[Import | ImportFrom] = []
    ) -> Custom:
        """
        Creates an intermediate node for a value with a custom representation which can be used as a result for your customization function.

        `create_code('{value-1!r}+1')` becomes `4+1` in the code.
        Use this when you need to control the exact string representation of a value.

        Arguments:
            code: Custom string representation to evaluate. This is required and will be evaluated using the snapshot context.
            imports: Optional list of Import and ImportFrom objects to add required imports to the generated code.
                     Example: `imports=[Import("os"), ImportFrom("pathlib", "Path")]`
        """
        import_vars = None

        # Try direct variable lookup for simple identifiers (fastest)
        if code.isidentifier():
            # Direct lookup with proper precedence: local > import > global
            if code in self._local_vars:
                return CustomCode(self._local_vars[code], code, imports)

            # Build import vars only if needed
            import_vars = self._build_import_vars(imports)
            if code in import_vars:
                return CustomCode(import_vars[code], code, imports)

            if code in self._global_vars:
                return CustomCode(self._global_vars[code], code, imports)

        # Try ast.literal_eval for simple literals (fast and safe)
        try:
            import ast

            return CustomCode(ast.literal_eval(code), code, imports)
        except (ValueError, SyntaxError):
            # Fall back to eval with context for complex expressions
            # Build evaluation context with proper precedence: global < import < local
            if import_vars is None:
                import_vars = self._build_import_vars(imports)
            eval_context = {
                **self._global_vars,
                **import_vars,
                **self._local_vars,
            }
            return CustomCode(eval(code, eval_context), code, imports)



import ast
from typing import Any
from typing import Iterator

from inline_snapshot._code_repr import mock_repr
from inline_snapshot._compare_context import compare_only
from inline_snapshot._customize._builder import Builder
from inline_snapshot._customize._custom import Custom
from inline_snapshot._customize._custom_call import CustomCall
from inline_snapshot._customize._custom_call import CustomDefault
from inline_snapshot._customize._custom_code import CustomCode
from inline_snapshot._customize._custom_dict import CustomDict
from inline_snapshot._customize._custom_sequence import CustomList
from inline_snapshot._customize._custom_sequence import CustomTuple
from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._customize._custom_unmanaged import CustomUnmanaged
from inline_snapshot._new_adapter import NewAdapter
from inline_snapshot._new_adapter import warn_star_expression
from inline_snapshot._unmanaged import is_unmanaged

from .._adapter_context import AdapterContext
from .._change import ChangeBase
from .generic_value import GenericValue


class AstToCustom:
    context: AdapterContext

    def __init__(self, context):
        self.eval = context.eval
        self.context = context

    def convert(self, value: Any, node: ast.expr):
        if is_unmanaged(value):
            return CustomUnmanaged(value)

        if warn_star_expression(node, self.context):
            return self.convert_generic(value, node)

        t = type(node).__name__
        return getattr(self, "convert_" + t, self.convert_generic)(value, node)

    def eval_convert(self, node):
        return self.convert(self.eval(node), node)

    def convert_generic(self, value: Any, node: ast.expr):
        if value is ...:
            return CustomUndefined()
        else:
            return CustomCode(value, ast.unparse(node))

    def convert_Call(self, value: Any, node: ast.Call):
        return CustomCall(
            self.eval_convert(node.func),
            [self.eval_convert(a) for a in node.args],
            {kw.arg: self.eval_convert(kw.value) for kw in node.keywords if kw.arg},
        )

    def convert_List(self, value: list, node: ast.List):

        return CustomList([self.convert(v, n) for v, n in zip(value, node.elts)])

    def convert_Tuple(self, value: tuple, node: ast.Tuple):
        return CustomTuple([self.convert(v, n) for v, n in zip(value, node.elts)])

    def convert_Dict(self, value: dict, node: ast.Dict):
        return CustomDict(
            {
                self.convert(k, k_node): self.convert(v, v_node)
                for (k, v), k_node, v_node in zip(value.items(), node.keys, node.values)
                if k_node is not None
            }
        )


class ValueToCustom:
    """this implementation is for cpython <= 3.10 only.
    It works similar to AstToCustom but can not handle calls
    """

    context: AdapterContext

    def __init__(self, context):
        self.context = context

    def convert(self, value: Any):
        if is_unmanaged(value):
            return CustomUnmanaged(value)

        if isinstance(value, CustomDefault):
            return self.convert(value.value)

        t = type(value).__name__
        return getattr(self, "convert_" + t, self.convert_generic)(value)

    def convert_generic(self, value: Any) -> Custom:
        if value is ...:
            return CustomUndefined()
        else:
            with mock_repr(self.context):
                result = Builder(self.context, _recursive=False)._get_handler(value)
            if isinstance(result, CustomCall) and result.function == type(value):
                function = self.convert(result.function)
                posonly_args = [self.convert(arg) for arg in result.args]
                kwargs = {k: self.convert(arg) for k, arg in result.kwargs.items()}

                return CustomCall(
                    function=function,
                    args=posonly_args,
                    kwargs=kwargs,
                )
            return CustomCode(value, "<unknown>")

    def convert_list(self, value: list):
        return CustomList([self.convert(v) for v in value])

    def convert_tuple(self, value: tuple):
        return CustomTuple([self.convert(v) for v in value])

    def convert_dict(self, value: dict):
        return CustomDict(
            {self.convert(k): self.convert(v) for (k, v) in value.items()}
        )


class UndecidedValue(GenericValue):
    def __init__(self, old_value, ast_node, context: AdapterContext):
        self._context = context
        self._ast_node = ast_node

        old_value = self.value_to_custom(old_value)

        assert isinstance(old_value, Custom)
        self._old_value = old_value
        self._new_value = CustomUndefined()

    def _change(self, cls):
        self.__class__ = cls

    def _new_code(self):
        assert False

    def _get_changes(self) -> Iterator[ChangeBase]:
        assert isinstance(self._new_value, CustomUndefined)

        new_value = self.to_custom(self._old_value._eval())

        adapter = NewAdapter(self._context)

        for change in adapter.compare(self._old_value, self._ast_node, new_value):
            assert change.flag == "update", change
            yield change

    def __eq__(self, other):
        if compare_only():
            return False

        from .._snapshot.eq_value import EqValue

        self._change(EqValue)
        return self == other

    def __le__(self, other):
        from .._snapshot.min_max_value import MinValue

        self._change(MinValue)
        return self <= other

    def __ge__(self, other):
        from .._snapshot.min_max_value import MaxValue

        self._change(MaxValue)
        return self >= other

    def __contains__(self, item):
        from .._snapshot.collection_value import CollectionValue

        self._change(CollectionValue)
        return item in self

    def __getitem__(self, item):
        from .._snapshot.dict_value import DictValue

        self._change(DictValue)
        return self[item]



from difflib import unified_diff
from itertools import islice
from pathlib import Path

from rich.syntax import Syntax


def diff(original_text, new_text):

    diff = "\n".join(
        islice(
            unified_diff(original_text.splitlines(), new_text.splitlines()),
            2,
            None,
        )
    ).strip()

    return Syntax(diff, "diff", theme="ansi_light", word_wrap=True)


def hexdump(bytes, style=False):
    from binascii import hexlify

    result = []
    for i in range(0, len(bytes), 16):
        line = bytes[i : i + 16]
        if style:
            dot = "[white].[/white]"
        else:
            dot = "."

        s = "".join(chr(c) if 32 <= c < 127 else dot for c in line)
        if style:
            result.append(
                f"[dark_green]{i:08x}[/]: [bright_black]{hexlify(line,' ',-2).decode():<39}[/] |[orange3]{s}{' '*(16-len(line))}[/]|"
            )
        else:
            result.append(f"{i:08x}: {hexlify(line,' ',-2).decode():<39} |{s:<16}|")

    return "\n".join(result)


class TextDiff:
    def rich_diff(self, original: Path, new: Path):

        return diff(original.read_text("utf-8"), new.read_text("utf-8"))

    def rich_show(self, path: Path):
        return Syntax.from_path(str(path), theme="ansi_light", word_wrap=True)


class BinaryDiff:
    def rich_diff(self, original: Path, new: Path):

        original_bytes = original.read_bytes()
        new_bytes = new.read_bytes()

        return diff(hexdump(original_bytes), hexdump(new_bytes))

    def rich_show(self, path: Path):
        content = path.read_bytes()

        if len(content) > 20 * 16:
            return f"<binary file ({len(content)} bytes)>"
        else:
            return hexdump(content, style=True)



import ast
from typing import Generator
from typing import Iterator
from typing import Union

from inline_snapshot._customize._custom_sequence import CustomList
from inline_snapshot._customize._custom_undefined import CustomUndefined

from .._change import ChangeBase
from .._change import Delete
from .._change import ListInsert
from .._change import Replace
from .._global_state import state
from .generic_value import GenericValue
from .generic_value import ignore_old_value


class CollectionValue(GenericValue):
    _current_op = "x in snapshot"
    _ast_node: Union[ast.List, ast.Tuple]
    _new_value: CustomList

    def __contains__(self, item):
        if isinstance(self._old_value, CustomUndefined):
            state().missing_values += 1

        if isinstance(self._new_value, CustomUndefined):
            self._new_value = CustomList([self.to_custom(item)])
        else:
            if item not in self._new_value._eval():
                self._new_value.value.append(self.to_custom(item))

        if ignore_old_value() or isinstance(self._old_value, CustomUndefined):
            return True
        else:
            return self._return(item in self._old_value._eval())

    def _new_code(self) -> Generator[ChangeBase, None, str]:
        code = yield from self._new_value._code_repr(self._context)
        return code

    def _get_changes(self) -> Iterator[ChangeBase]:
        assert isinstance(self._old_value, CustomList), self._old_value
        assert isinstance(self._new_value, CustomList), self._new_value

        if self._ast_node is None:
            elements = [None] * len(self._old_value.value)
        else:
            assert isinstance(self._ast_node, ast.List)
            elements = self._ast_node.elts

        for old_value, old_node in zip(self._old_value.value, elements):
            if old_value not in self._new_value.value:
                yield Delete(
                    flag="trim",
                    file=self._file,
                    node=old_node,
                    old_value=old_value,
                )
                continue

            # check for update
            new_code = yield from self.to_custom(old_value._eval())._code_repr(
                self._context
            )

            if self._file.code_changed(old_node, new_code):

                yield Replace(
                    node=old_node,
                    file=self._file,
                    new_code=new_code,
                    flag="update",
                    old_value=old_value,
                    new_value=old_value,
                )

        new_codes = []
        new_values = []
        for v in self._new_value.value:
            if v not in self._old_value.value:
                new_code = yield from v._code_repr(self._context)
                new_codes.append(new_code)
                new_values.append(v._eval())

        if new_codes:
            yield ListInsert(
                flag="fix",
                file=self._file,
                node=self._ast_node,
                position=len(self._old_value.value),
                new_code=new_codes,
                new_values=new_values,
            )



import ast
import inspect
from typing import Any
from typing import Iterator
from typing import TypeVar

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._customize._custom_undefined import CustomUndefined
from inline_snapshot._generator_utils import with_flag
from inline_snapshot._types import SnapshotRefBase

from ._change import CallArg
from ._change import ChangeBase
from ._global_state import state
from ._sentinels import undefined
from ._snapshot.undecided_value import UndecidedValue


class ReprWrapper:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return self.func.__name__


_T = TypeVar("_T")


def repr_wrapper(func: _T) -> _T:
    return ReprWrapper(func)  # type: ignore


def create_snapshot(Type, obj, extra_frames=0):

    frame = inspect.currentframe()
    assert frame is not None
    frame = frame.f_back
    assert frame is not None
    frame = frame.f_back
    assert frame is not None

    for _ in range(extra_frames):
        frame = frame.f_back
        assert frame is not None

    context = AdapterContext(frame)

    if not state().active:
        if obj is undefined:
            raise AssertionError(
                "your snapshot is missing a value run pytest with --inline-snapshot=create"
            )
        else:
            return Type.create_raw(obj, context)

    Type.check_context(context)

    key = id(frame.f_code), frame.f_lasti

    if key not in state().snapshots:
        node = context.expr.node
        if node is None:
            # we can run without knowing of the calling expression but we will not be able to fix code
            new = Type(obj, None, context)
        else:
            assert isinstance(node, ast.Call)
            new = Type(obj, context.expr, context)
        state().snapshots[key] = new
    else:
        new = state().snapshots[key]
        new._re_eval(obj, context)

    return new.result()


@repr_wrapper
def snapshot(obj: Any = undefined) -> Any:
    """`snapshot()` is a placeholder for some value.

    `pytest --inline-snapshot=create` will create the value which matches your conditions.

    >>> assert 5 == snapshot()
    >>> assert 5 <= snapshot()
    >>> assert 5 >= snapshot()
    >>> assert 5 in snapshot()

    `snapshot()[key]` can be used to create sub-snapshots.

    The generated value will be inserted as argument to `snapshot()`

    >>> assert 5 == snapshot(5)

    `snapshot(value)` has general the semantic of an noop which returns `value`.
    """

    return create_snapshot(SnapshotReference, obj, 1)


class SnapshotReference(SnapshotRefBase):
    def __init__(self, value, expr, context: AdapterContext):
        self._expr = expr
        node = expr.node.args[0] if expr is not None and expr.node.args else None
        self._value = UndecidedValue(value, node, context)
        self._context = context

    def result(self):
        return self._value

    @staticmethod
    def create_raw(obj, context: AdapterContext):
        return obj

    def _changes(self) -> Iterator[ChangeBase]:

        if (
            isinstance(self._value._old_value, CustomUndefined)
            if self._expr is None
            else not self._expr.node.args
        ):

            if isinstance(self._value._new_value, CustomUndefined):
                return

            new_code = yield from with_flag(self._value._new_code(), "create")

            yield CallArg(
                flag="create",
                file=self._value._file,
                node=self._expr.node if self._expr is not None else None,
                arg_pos=0,
                arg_name=None,
                new_code=new_code,
                new_value=self._value._new_value,
            )

        else:

            yield from self._value._get_changes()

    def _re_eval(self, obj, context: AdapterContext):
        self._value._re_eval(obj, context)



from __future__ import annotations

import ast
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import DefaultDict
from typing import Tuple
from typing import cast

from asttokens.util import Token
from executing.executing import EnhancedAST

from inline_snapshot._external._external_location import Location
from inline_snapshot._external._find_external import ensure_import
from inline_snapshot._source_file import SourceFile

from ._rewrite_code import ChangeRecorder
from ._rewrite_code import end_of
from ._rewrite_code import start_of

if TYPE_CHECKING:
    from inline_snapshot._external._format._protocol import Format


class ChangeBase:
    flag: str

    def rich_diff(self):
        raise NotImplementedError

    def apply_external_changes(self):
        raise NotImplementedError

    def apply(self, recorder: ChangeRecorder):
        raise NotImplementedError


@dataclass()
class ExternalChange(ChangeBase):
    flag: str

    new_file: Path

    old_location: Location
    new_location: Location

    format: Format

    def rich_diff(self):

        title = str(self.new_location)
        if title != (old_name := str(self.old_location)) and self.old_location.stem:
            title = f"{old_name} -> {title}"

        if (
            self.old_location.exists()
            and self.old_location.suffix == self.new_location.suffix
        ):
            with self.old_location.load() as old_file:
                return title, self.format.rich_diff(old_file, self.new_file)
        else:
            return title, self.format.rich_show(self.new_file)

    def apply_external_changes(self):
        self.new_location.store(self.new_file)

    def apply(self, recorder: ChangeRecorder):
        pass


@dataclass()
class ExternalRemove(ChangeBase):
    flag: str

    old_location: Location

    def rich_diff(self):

        title = f"delete {self.old_location}"

        from inline_snapshot._external._format._protocol import (
            get_format_handler_from_suffix,
        )

        with self.old_location.load() as old_file:
            assert self.old_location.suffix
            format = get_format_handler_from_suffix(self.old_location.suffix)
            return title, format.rich_show(old_file)

    def apply_external_changes(self):
        self.old_location.delete()

    def apply(self, recorder: ChangeRecorder):
        pass


@dataclass()
class Change(ChangeBase):
    flag: str
    file: SourceFile

    def rich_diff(self):
        return None

    @property
    def filename(self):
        return self.file.filename

    def apply(self, recorder: ChangeRecorder):
        raise NotImplementedError()

    def apply_external_changes(self):
        pass


@dataclass()
class RequiredImport(Change):
    module: str
    name: str | None = None


@dataclass()
class Delete(Change):
    node: ast.AST | None
    old_value: Any


@dataclass()
class ListInsert(Change):
    node: ast.List | ast.Tuple
    position: int

    new_code: list[str]
    new_values: list[Any]

    def __post_init__(self):
        self.new_code = [self.file.format_expression(v) for v in self.new_code]


@dataclass()
class DictInsert(Change):
    node: ast.Dict
    position: int

    new_code: list[tuple[str, str]]
    new_values: list[tuple[Any, Any]]

    def __post_init__(self):
        self.new_code = [
            (self.file.format_expression(k), self.file.format_expression(v))
            for k, v in self.new_code
        ]


@dataclass()
class Replace(Change):
    node: ast.AST

    new_code: str
    old_value: Any
    new_value: Any

    def apply(self, recorder: ChangeRecorder):
        change = recorder.new_change()
        range = self.file.asttokens().get_text_positions(self.node, False)
        change.replace(range, self.new_code, filename=self.filename)

    def __post_init__(self):
        self.new_code = self.file.format_expression(self.new_code)


@dataclass()
class CallArg(Change):
    node: ast.Call | None
    arg_pos: int | None
    arg_name: str | None

    new_code: str
    new_value: Any

    def __post_init__(self):
        self.new_code = self.file.format_expression(self.new_code)


TokenRange = Tuple[Token, Token]


def brace_tokens(source, node) -> TokenRange:
    first_token, *_, end_token = source.asttokens().get_tokens(node)
    return first_token, end_token


def generic_sequence_update(
    source: SourceFile,
    parent: ast.List | ast.Tuple | ast.Dict | ast.Call,
    brace_tokens: TokenRange,
    parent_elements: list[TokenRange | None],
    to_insert: dict[int, list[str]],
    recorder: ChangeRecorder,
):
    rec = recorder.new_change()

    new_code = []
    deleted = False
    last_token, end_token = brace_tokens
    is_start = True
    elements = 0

    for index, entry in enumerate(parent_elements):
        if index in to_insert:
            new_code += to_insert[index]
        if entry is None:
            deleted = True
        else:
            first_token, new_last_token = entry
            elements += len(new_code) + 1

            if deleted or new_code:

                code = ""
                if new_code:
                    code = ", ".join(new_code) + ", "
                if not is_start:
                    code = ", " + code

                rec.replace(
                    (end_of(last_token), start_of(first_token)),
                    code,
                    filename=source.filename,
                )
            new_code = []
            deleted = False
            last_token = new_last_token
            is_start = False

    for i, insert_elements in sorted(to_insert.items()):
        if i >= len(parent_elements):
            new_code += insert_elements
            elements += len(new_code)

    if new_code or deleted or elements == 1 or len(parent_elements) <= 1:
        code = ", ".join(new_code)
        if not is_start and code:
            code = ", " + code

        if elements == 1 and isinstance(parent, ast.Tuple):
            # trailing comma for tuples (1,)
            code += ","

        rec.replace(
            (end_of(last_token), start_of(end_token)),
            code,
            filename=source.filename,
        )


def apply_all(all_changes: list[ChangeBase], recorder: ChangeRecorder):
    by_parent: dict[EnhancedAST, list[Delete | DictInsert | ListInsert | CallArg]] = (
        defaultdict(list)
    )
    sources: dict[EnhancedAST, SourceFile] = {}

    # file -> module -> names
    imports_by_file: dict[str, dict[str, set]] = defaultdict(lambda: defaultdict(set))
    module_imports_by_file: dict[str, set] = defaultdict(set)

    for change in all_changes:
        if isinstance(change, Delete):
            node = cast(EnhancedAST, change.node).parent
            if isinstance(node, ast.keyword):
                node = node.parent
            by_parent[node].append(change)
            sources[node] = change.file

        elif isinstance(change, (DictInsert, ListInsert, CallArg)):
            node = cast(EnhancedAST, change.node)
            by_parent[node].append(change)
            sources[node] = change.file
        elif isinstance(change, RequiredImport):
            if change.name:
                imports_by_file[change.file.filename][change.module].add(change.name)
            else:
                module_imports_by_file[change.file.filename].add(change.module)
        else:
            change.apply(recorder)

    for filename in set(imports_by_file) | set(module_imports_by_file):
        imports = imports_by_file.get(filename, defaultdict(set))
        module_imports = module_imports_by_file.get(filename, set())
        ensure_import(filename, imports, module_imports, recorder)

    for parent, changes in by_parent.items():
        source = sources[parent]

        if isinstance(parent, (ast.List, ast.Tuple)):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            to_insert = {
                change.position: change.new_code
                for change in changes
                if isinstance(change, ListInsert)
            }

            def list_token_range(entry):
                r = list(source.asttokens().get_tokens(entry))
                return r[0], r[-1]

            generic_sequence_update(
                source,
                parent,
                brace_tokens(source, parent),
                [None if e in to_delete else list_token_range(e) for e in parent.elts],
                to_insert,
                recorder,
            )

        elif isinstance(parent, ast.Call):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            atok = source.asttokens()

            def arg_token_range(node):
                if isinstance(node.parent, ast.keyword):
                    node = node.parent
                r = list(atok.get_tokens(node))
                return r[0], r[-1]

            braces_left = atok.next_token(list(atok.get_tokens(parent.func))[-1])
            assert braces_left.string == "("
            braces_right = list(atok.get_tokens(parent))[-1]
            assert braces_right.string == ")"

            to_insert = DefaultDict(list)

            other_call_args: list[CallArg] = []
            call_args: list[CallArg] = []

            for change in changes:
                if isinstance(change, CallArg):
                    call_args.append(change)

            max_pos = len(parent.args)

            for change in sorted(call_args, key=lambda arg: arg.arg_pos or 0):
                if change.arg_name is None:
                    assert change.arg_pos is not None
                    to_insert[min(change.arg_pos, max_pos)].append(change.new_code)
                else:
                    other_call_args.append(change)

            end_pos = (
                max([len(parent.args) + len(parent.keywords), *to_insert.keys()]) + 1
            )

            for change in other_call_args:
                position = change.arg_pos if change.arg_pos is not None else end_pos
                to_insert[position].append(f"{change.arg_name}={change.new_code}")

            generic_sequence_update(
                source,
                parent,
                (braces_left, braces_right),
                [
                    None if e in to_delete else arg_token_range(e)
                    for e in parent.args + [kw.value for kw in parent.keywords]
                ],
                to_insert,
                recorder,
            )

        elif isinstance(parent, ast.Dict):
            to_delete = {
                change.node for change in changes if isinstance(change, Delete)
            }
            to_insert = {
                change.position: [f"{key}: {value}" for key, value in change.new_code]
                for change in changes
                if isinstance(change, DictInsert)
            }

            def dict_token_range(key, value):
                return (
                    list(source.asttokens().get_tokens(key))[0],
                    list(source.asttokens().get_tokens(value))[-1],
                )

            generic_sequence_update(
                source,
                parent,
                brace_tokens(source, parent),
                [
                    None if value in to_delete else dict_token_range(key, value)
                    for key, value in zip(parent.keys, parent.values)
                ],
                to_insert,
                recorder,
            )

        else:
            assert False, parent



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Generator

from inline_snapshot._adapter_context import AdapterContext
from inline_snapshot._change import ChangeBase

from ._custom import Custom


@dataclass()
class CustomUnmanaged(Custom):
    value: Any

    def _code_repr(
        self, context: AdapterContext
    ) -> Generator[ChangeBase, None, str]:  # pragma: no cover
        yield from ()
        return "'unmanaged'"

    def _map(self, f):
        return f(self.value)



from pathlib import Path
from typing import Optional
from typing import Union
from typing import cast

from inline_snapshot._external._external_location import FileLocation
from inline_snapshot._external._format._protocol import Format
from inline_snapshot._external._format._protocol import get_format_handler_from_suffix
from inline_snapshot._external._storage._protocol import StorageLookupError
from inline_snapshot._global_state import state
from inline_snapshot._types import SnapshotRefBase

from ._external_base import ExternalBase


class ExternalFile(ExternalBase, SnapshotRefBase):

    def __init__(self, filename: Path, format: Format):
        self._filename = filename
        self._format = format
        self._tmp_file = None

        self._location = FileLocation(self._filename)
        self._original_location = self._location
        super().__init__()

    def _is_empty(self):
        return not self._filename.exists()

    def _assign(self, other):

        self._tmp_file = state().new_tmp_path(self._location.suffix)
        self._format.encode(other, self._tmp_file)

    def __repr__(self):
        return f"external_file({str(self._filename)!r})"

    def _load_value(self):
        try:
            return self._format.decode(self._filename)
        except FileNotFoundError:
            raise StorageLookupError(
                f"can not read {self._filename}", files=[self._filename]
            )


def external_file(path: Union[Path, str], *, format: Optional[str] = None):
    """
    Arguments:
        path: the path to the external file, relative to the directory of the current file.
        format: overwrite the format handler which should be used to load and save the content.
                It can be used to treat markdown files as text files with `format=".txt"` for example.
    """
    path = Path(path)

    if not path.is_absolute():
        from inspect import currentframe

        frame = currentframe()
        assert frame
        frame = frame.f_back
        assert frame
        path = Path(frame.f_code.co_filename).parent / path

    path = path.resolve()

    if format is None:
        format = path.suffix

    format_handler = get_format_handler_from_suffix(format)

    if not state().active:
        return format_handler.decode(path)

    key = ("file", path)
    if key not in state().snapshots:
        new = ExternalFile(path, format_handler)
        state().snapshots[key] = new
    else:
        new = cast(ExternalFile, state().snapshots[key])

    assert new._format.suffix == format_handler.suffix

    return new



import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional

from inline_snapshot._exceptions import UsageError

if sys.version_info >= (3, 11):
    from tomllib import loads
else:
    from tomli import loads


@dataclass
class Config:
    hash_length: int = 12
    default_flags: List[str] = field(default_factory=lambda: ["short-report"])
    default_flags_tui: List[str] = field(default_factory=lambda: ["short-report"])
    shortcuts: Dict[str, List[str]] = field(default_factory=dict)
    format_command: str = ""
    storage_dir: Optional[Path] = None
    show_updates: bool = False
    test_directories: Optional[List[Path]] = None
    default_storage: str = "uuid"


def read_config(path: Path, config=Config()) -> Config:
    tool_config = {}

    if path.exists():
        data = loads(path.read_text("utf-8"))

        try:
            tool_config = data["tool"]["inline-snapshot"]
        except KeyError:
            pass

    try:
        config.hash_length = tool_config["hash-length"]
    except KeyError:
        pass

    try:
        config.default_flags = tool_config["default-flags"]
    except KeyError:
        pass

    try:
        config.default_flags_tui = tool_config["default-flags-tui"]
    except KeyError:
        pass

    try:
        config.show_updates = tool_config["show-updates"]
    except KeyError:
        pass

    config.shortcuts = tool_config.get(
        "shortcuts", {"fix": ["create", "fix"], "review": ["review"]}
    )

    def to_path(pathname: str) -> Path:
        result = Path(pathname)
        if not result.is_absolute():
            # Make it relative to pyproject.toml, and absolute.
            result = path.parent.joinpath(result).absolute()
        return result.resolve()

    config.storage_dir = (
        to_path(name) if (name := tool_config.get("storage-dir")) else None
    )

    test_directories = tool_config.get("test-dir")

    if isinstance(test_directories, str):
        test_directories = [test_directories]

    if isinstance(test_directories, list):
        config.test_directories = [to_path(d) for d in test_directories]
    elif (
        test_directories is None
        and path.exists()
        and (std_tests := path.parent / "tests").exists()
        and std_tests.is_dir()
    ):
        config.test_directories = [std_tests.resolve()]
    elif test_directories is None:
        config.test_directories = None
    else:
        raise UsageError(f"test-dir has to be a directory or list of directories")

    config.default_storage = tool_config.get("default-storage", "uuid")

    if config.default_storage not in ("uuid", "hash"):
        raise UsageError(
            f'default-storage has to be uuid or hash but is "{config.default_storage}"'
        )

    config.format_command = tool_config.get("format-command", "")

    return config



import subprocess as sp
import warnings

from rich.markup import escape

from ._problems import raise_problem


def enforce_formatting():
    from inline_snapshot._global_state import state

    return bool(state().config.format_command)


def file_mode_for_path(path):
    from black import FileMode
    from black import find_pyproject_toml
    from black import parse_pyproject_toml

    mode = FileMode()
    pyproject_path = find_pyproject_toml((), path)
    if pyproject_path is not None:
        config = parse_pyproject_toml(pyproject_path)

        if "line_length" in config:
            mode.line_length = int(config["line_length"])
        if "skip_magic_trailing_comma" in config:
            mode.magic_trailing_comma = not config["skip_magic_trailing_comma"]
        if "skip_string_normalization" in config:
            # The ``black`` command line argument is
            # ``--skip-string-normalization``, but the parameter for
            # ``black.Mode`` needs to be the opposite boolean of
            # ``skip-string-normalization``, hence the inverse boolean
            mode.string_normalization = not config["skip_string_normalization"]
        if "preview" in config:
            mode.preview = config["preview"]

    return mode


def format_code(text, filename):
    from inline_snapshot._global_state import state

    if state().config.format_command:
        format_command = state().config.format_command.format(filename=filename)

        # Split the command at | to handle pipelines properly
        # https://github.com/15r10nk/inline-snapshot/issues/320
        commands = [cmd.strip() for cmd in format_command.split("|")]

        current_input = text.encode("utf-8")

        for command in commands:
            result = sp.run(
                command, shell=True, input=current_input, capture_output=True
            )
            if result.returncode != 0:
                raise_problem(
                    f"""\
[b]The format_command '{escape(format_command)}' caused the following error:[/b]
"""
                    + result.stdout.decode("utf-8")
                    + result.stderr.decode("utf-8")
                )
                return text
            current_input = result.stdout

        return current_input.decode("utf-8")

    try:
        from black import format_str
    except ImportError:
        raise_problem(
            f"""\
[b]inline-snapshot is not able to format your code.[/b]
This issue can be solved by:
 * installing {escape('inline-snapshot[black]')} which gives you the same formatting like in older versions
 * adding a `format-command` to your pyproject.toml (see [link=https://15r10nk.github.io/inline-snapshot/latest/configuration/#format-command]https://15r10nk.github.io/inline-snapshot/latest/configuration/#format-command[/link] for more information).
"""
        )
        return text

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        mode = file_mode_for_path(filename)

        try:
            # "a\n" is a work around for https://github.com/15r10nk/inline-snapshot/issues/301
            # it prevents that " " gets treated as a docstring and gets striped by black.
            return format_str("a\n" + text, mode=mode)[2:].lstrip()
        except:
            raise_problem(
                """\
[b]black could not format your code, which might be caused by this issue:[/b]
    [link=https://github.com/15r10nk/inline-snapshot/issues/138]https://github.com/15r10nk/inline-snapshot/issues/138[/link]\
"""
            )
            return text



from typing import Generator
from typing import Iterator

from inline_snapshot._customize._custom_undefined import CustomUndefined

from .._change import ChangeBase
from .._change import Replace
from .._global_state import state
from .generic_value import GenericValue
from .generic_value import ignore_old_value


class MinMaxValue(GenericValue):
    """Generic implementation for <=, >="""

    @staticmethod
    def cmp(a, b):
        raise NotImplementedError

    def _generic_cmp(self, other):
        if isinstance(self._old_value, CustomUndefined):
            state().missing_values += 1

        if isinstance(self._new_value, CustomUndefined):
            self._new_value = self.to_custom(other)
            if isinstance(self._old_value, CustomUndefined) or ignore_old_value():
                return True
            return self._return(self.cmp(self._old_value._eval(), other))
        else:
            if not self.cmp(self._new_value._eval(), other):
                self._new_value = self.to_custom(other)

        return self._return(self.cmp(self._visible_value()._eval(), other))

    def _new_code(self) -> Generator[ChangeBase, None, str]:
        code = yield from self._new_value._code_repr(self._context)
        return code

    def _get_changes(self) -> Iterator[ChangeBase]:
        new_code = yield from self._new_code()

        if not self.cmp(self._old_value._eval(), self._new_value._eval()):
            flag = "fix"
        elif not self.cmp(self._new_value._eval(), self._old_value._eval()):
            flag = "trim"
        elif self._file.code_changed(self._ast_node, new_code):
            flag = "update"
        else:
            return

        yield Replace(
            node=self._ast_node,
            file=self._file,
            new_code=new_code,
            flag=flag,
            old_value=self._old_value._eval(),
            new_value=self._new_value._eval(),
        )


class MinValue(MinMaxValue):
    """
    handles:

    >>> snapshot(5) <= 6
    True

    >>> 6 >= snapshot(5)
    True

    """

    _current_op = "x >= snapshot"

    @staticmethod
    def cmp(a, b):
        return a <= b

    __le__ = MinMaxValue._generic_cmp


class MaxValue(MinMaxValue):
    """
    handles:

    >>> snapshot(5) >= 4
    True

    >>> 4 <= snapshot(5)
    True

    """

    _current_op = "x <= snapshot"

    @staticmethod
    def cmp(a, b):
        return a >= b

    __ge__ = MinMaxValue._generic_cmp



from __future__ import annotations

from typing import Set
from typing import cast

from ._types import Category


class Flags:
    """
    fix: the value needs to be changed to pass the tests
    update: the value should be updated because the token-stream has changed
    create: the snapshot is empty `snapshot()`
    trim: the snapshot contains more values than necessary. 1 could be trimmed in `5 in snapshot([1,5])`.
    """

    def __init__(self, flags: set[Category] = set()):
        self.create = "create" in flags
        self.fix = "fix" in flags
        self.fix_assert = "fix-assert" in flags
        self.trim = "trim" in flags
        self.update = "update" in flags

    def to_set(self) -> set[Category]:
        return cast(
            Set[Category], {k.replace("_", "-") for k, v in self.__dict__.items() if v}
        )

    def __iter__(self):
        return (k.replace("_", "-") for k, v in self.__dict__.items() if v)

    def __repr__(self):
        return f"Flags({self.to_set()})"

    @staticmethod
    def all() -> Flags:
        return Flags({"fix", "create", "update", "trim", "fix-assert"})
